"""
Core OpenBCI object for handling connections and samples from the WiFi Shield
Note that the LIB will take care on its own to print incoming ASCII messages if any (FIXME, BTW).
EXAMPLE USE:
def handle_sample(sample):
  print(sample.channels_data)
wifi = OpenBCIWifi()
wifi.start_stream(handle_sample)
asyncore.loop()

TODO: Cyton/Ganglion JSON
TODO: Ganglion Raw
TODO: Cyton Raw
"""
from __future__ import print_function
import asyncore
import atexit
import json
import logging
import re
import socket
import timeit
import time
import struct

try:
    import urllib2
except ImportError:
    import urllib

import requests
import xmltodict
# EXTRACTED FROM : www.github.com/pyopenBCI
import ssdp

SAMPLE_RATE = 0  # Hz

'''
#Commands for in SDK
command_stop = "s";
command_startBinary = "b";
'''


class OpenBCIWiFi(object):
    """
    Handle a connection to an OpenBCI wifi shield.
    Args:
      ip_address: The IP address of the WiFi Shield, "None" to attempt auto-detect.
      shield_name: The unique name of the WiFi Shield, such as `OpenBCI-2AD4`, will use SSDP to
        get IP address still, if `shield_name` is "None" and `ip_address` is "None",
        will connect to the first WiFi Shield found using SSDP
      sample_rate: The sample rate to set the attached board to. If the sample rate picked
        is not a sample rate the attached board can support, i.e. you send 300 to Cyton,
        then error will be thrown.
      log:
      timeout: in seconds, disconnect / reconnect after a period without new data
        should be high if impedance check
      max_packets_to_skip: will try to disconnect / reconnect after too many packets are skipped
    """

    def __init__(self, ip_address=None, shield_name=None, sample_rate=None, log=True, timeout=3,
                 max_packets_to_skip=20, latency=10000, high_speed=True, ssdp_attempts=5,
                 num_channels=8, local_ip_address=None, local_port=None):
        # these one are used
        self.daisy = False
        self.gains = None
        self.high_speed = high_speed
        self.impedance = False
        self.ip_address = ip_address
        self.latency = latency
        self.log = log  # print_incoming_text needs log
        self.max_packets_to_skip = max_packets_to_skip
        self.num_channels = num_channels
        self.sample_rate = sample_rate
        self.shield_name = shield_name
        self.ssdp_attempts = ssdp_attempts
        self.streaming = False
        self.timeout = timeout

        # might be handy to know API
        self.board_type = "none"
        # number of EEG channels
        self.eeg_channels_per_sample = 0
        self.read_state = 0
        self.log_packet_count = 0
        self.packets_dropped = 0
        self.time_last_packet = 0

        if self.log:
            print("Welcome to OpenBCI Native WiFi Shield Driver - Please contribute code!")

        self.local_ip_address = local_ip_address
        if not self.local_ip_address:
            self.local_ip_address = self._get_local_ip_address()
        self.local_port=local_port
        if not self.local_port:
            # Intentionally bind to port 0 -- so auto allocate unused port
            self.local_port=0

        self.local_wifi_server = WiFiShieldServer(self.local_ip_address, self.local_port)
        self.local_wifi_server_port = self.local_wifi_server.socket.getsockname()[1]
        if self.log:
            print("Opened socket on %s:%d" %
                  (self.local_ip_address, self.local_wifi_server_port))

        if ip_address is None:
            for i in range(ssdp_attempts):
                try:
                    self.find_wifi_shield(wifi_shield_cb=self.on_shield_found)
                    break
                except OSError:
                    # Try again
                    if self.log:
                        print("Did not find any WiFi Shields")
        else:
            self.on_shield_found(ip_address)

    def on_shield_found(self, ip_address):
        self.ip_address = ip_address
        self.connect()
        # Disconnects from board when terminated
        atexit.register(self.disconnect)

    def loop(self):
        asyncore.loop()

    def _get_local_ip_address(self):
        """
        Gets the local ip address of this computer
        @returns str Local IP address
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        local_ip_address = s.getsockname()[0]
        s.close()
        return local_ip_address

    def getBoardType(self):
        """ Returns the version of the board """
        return self.board_type

    def setImpedance(self, flag):
        """ Enable/disable impedance measure """
        self.impedance = bool(flag)

    def connect(self):
        """ Connect to the board and configure it. Note: recreates various objects upon call. """
        if self.ip_address is None:
            raise ValueError('self.ip_address cannot be None')

        if self.log:
            print("Init WiFi connection with IP: " + self.ip_address)

        """
        Docs on these HTTP requests and more are found:
        https://app.swaggerhub.com/apis/pushtheworld/openbci-wifi-server/1.3.0
        """

        res_board = requests.get("http://%s/board" % self.ip_address)

        if res_board.status_code == 200:
            board_info = res_board.json()
            if not board_info['board_connected']:
                raise RuntimeError("No board connected to WiFi Shield. "
                                   "To learn how to connect to a Cyton or Ganglion visit "
                                   "http://docs.openbci.com/Tutorials/03-Wifi_Getting_Started_Guide")
            self.board_type = board_info['board_type']
            self.eeg_channels_per_sample = board_info['num_channels']
            if self.log:
                print("Connected to %s with %s channels" %
                      (self.board_type, self.eeg_channels_per_sample))

        self.gains = None
        if self.board_type == 'cyton':
            self.gains = [24, 24, 24, 24, 24, 24, 24, 24]
            self.daisy = False
        elif self.board_type == 'daisy':
            self.gains = [24, 24, 24, 24, 24, 24, 24,
                          24, 24, 24, 24, 24, 24, 24, 24, 24]
            self.daisy = True
        elif self.board_type == 'ganglion':
            self.gains = [51, 51, 51, 51]
            self.daisy = False
        self.local_wifi_server.set_daisy(daisy=self.daisy)
        self.local_wifi_server.set_parser(
            ParseRaw(gains=self.gains, board_type=self.board_type))

        if self.high_speed:
            output_style = 'raw'
        else:
            output_style = 'json'
        res_tcp_post = requests.post("http://%s/tcp" % self.ip_address,
                                     json={
                                         'ip': self.local_ip_address,
                                         'port': self.local_wifi_server_port,
                                         'output': output_style,
                                         'delimiter': True,
                                         'latency': self.latency
                                     })
        if res_tcp_post.status_code == 200:
            tcp_status = res_tcp_post.json()
            if tcp_status['connected']:
                if self.log:
                    print("WiFi Shield to Python TCP Socket Established")
            else:
                raise RuntimeWarning("WiFi Shield is not able to connect to local server."
                                     "Please open an issue.")

    def init_streaming(self):
        """ Tell the board to record like crazy. """
        res_stream_start = requests.get(
            "http://%s/stream/start" % self.ip_address)
        if res_stream_start.status_code == 200:
            self.streaming = True
            self.packets_dropped = 0
            self.time_last_packet = timeit.default_timer()
        else:
            raise EnvironmentError("Unable to start streaming."
                                   "Check API for status code %d on /stream/start"
                                   % res_stream_start.status_code)

    def find_wifi_shield(self, shield_name=None, wifi_shield_cb=None):
        """Detects Ganglion board MAC address if more than 1, will select first. Needs root."""

        if self.log:
            print("Try to find WiFi shields on your local wireless network")
            print("Scanning for %d seconds nearby devices..." % self.timeout)

        list_ip = []
        list_id = []
        found_shield = False

        def wifi_shield_found(response):
            res = requests.get(response.location, verify=False).text
            device_description = xmltodict.parse(res)
            try : 
               cur_shield_name = str(
                  device_description['root']['device']['serialNumber'])
               cur_base_url = str(device_description['root']['URLBase'])
               cur_ip_address = re.findall(r'[0-9]+(?:\.[0-9]+){3}', cur_base_url)[0]
            except : 
               # it's probably a direct connection
               # TODO [] : check for a 404 response...
               cur_shield_name = 'DIRECT SHIELD'
               cur_ip_address = '192.168.4.1'
            list_id.append(cur_shield_name)
            list_ip.append(cur_ip_address)
            found_shield = True
            if shield_name is None:
                print("Found WiFi Shield %s with IP Address %s" %
                      (cur_shield_name, cur_ip_address))
                if wifi_shield_cb is not None:
                    wifi_shield_cb(cur_ip_address)
            else:
                if shield_name == cur_shield_name:
                    if wifi_shield_cb is not None:
                        wifi_shield_cb(cur_ip_address)

        ssdp_hits = ssdp.discover("urn:schemas-upnp-org:device:Basic:1", timeout=self.timeout,
                                  wifi_found_cb=wifi_shield_found)

        nb_wifi_shields = len(list_id)

        if nb_wifi_shields < 1:
            print("No WiFi Shields found ;(")
            raise OSError('Cannot find OpenBCI WiFi Shield with local name')

        if nb_wifi_shields > 1:
            print(
                "Found " + str(nb_wifi_shields) +
                ", selecting first named: " + list_id[0] +
                " with IPV4: " + list_ip[0])
            return list_ip[0]

    def write_command(self, output):
        """
        Pass through commands from the WiFi Shield to the Carrier board
        :param output:
        :return:
        """
        res_command_post = requests.post("http://%s/command" % self.ip_address,
                                         json={'command': output})
        if res_command_post.status_code == 200:
            ret_val = res_command_post.text
            if self.log:
                print(ret_val)
            return ret_val
        else:
            if self.log:
                print("Error code: %d %s" %
                      (res_command_post.status_code, res_command_post.text))
            raise RuntimeError("Error code: %d %s" % (
                res_command_post.status_code, res_command_post.text))

    def getSampleRate(self):
        return self.sample_rate

    def getNbEEGChannels(self):
        """Will not get new data on impedance check."""
        return self.eeg_channels_per_sample

    def start_stream(self, callback, lapse=-1):
        """
        Start handling streaming data from the board. Call a provided callback
        for every single sample that is processed
        Args:
          callback: A callback function, or a list of functions, that will receive a single
            argument of the OpenBCISample object captured.
        """
        start_time = timeit.default_timer()

        # Enclose callback function in a list if it comes alone
        if not isinstance(callback, list):
            self.local_wifi_server.set_callback(callback)
        else:
            self.local_wifi_server.set_callback(callback[0])

        if not self.streaming:
            self.init_streaming()

        # while self.streaming:
        #     # should the board get disconnected and we could not wait for notification anymore
        #     #  a reco should be attempted through timeout mechanism
        #     try:
        #         # at most we will get one sample per packet
        #         self.waitForNotifications(1. / self.getSampleRate())
        #     except Exception as e:
        #         print("Something went wrong while waiting for a new sample: " + str(e))
        #     # retrieve current samples on the stack
        #     samples = self.delegate.getSamples()
        #     self.packets_dropped = self.delegate.getMaxPacketsDropped()
        #     if samples:
        #         self.time_last_packet = timeit.default_timer()
        #         for call in callback:
        #             for sample in samples:
        #                 call(sample)
        #
        #     if (lapse > 0 and timeit.default_timer() - start_time > lapse):
        #         self.stop();
        #     if self.log:
        #         self.log_packet_count = self.log_packet_count + 1;
        #
        #     # Checking connection -- timeout and packets dropped
        #     self.check_connection()

    def test_signal(self, signal):
        """ Enable / disable test signal """
        if signal == 0:
            self.warn("Disabling synthetic square wave")
            try:
                self.write_command(']')
            except Exception as e:
                print("Something went wrong while setting signal: " + str(e))
        elif signal == 1:
            self.warn("Enabling synthetic square wave")
            try:
                self.write_command('[')
            except Exception as e:
                print("Something went wrong while setting signal: " + str(e))
        else:
            self.warn("%s is not a known test signal. Valid signal is 0-1" % signal)

    def set_channel(self, channel, toggle_position):
        """ Enable / disable channels """
        try:
            if channel > self.num_channels:
                raise ValueError('Cannot set non-existant channel')
            # Commands to set toggle to on position
            if toggle_position == 1:
                if channel is 1:
                    self.write_command('!')
                if channel is 2:
                    self.write_command('@')
                if channel is 3:
                    self.write_command('#')
                if channel is 4:
                    self.write_command('$')
                if channel is 5:
                    self.write_command('%')
                if channel is 6:
                    self.write_command('^')
                if channel is 7:
                    self.write_command('&')
                if channel is 8:
                    self.write_command('*')
                if channel is 9:
                    self.write_command('Q')
                if channel is 10:
                    self.write_command('W')
                if channel is 11:
                    self.write_command('E')
                if channel is 12:
                    self.write_command('R')
                if channel is 13:
                    self.write_command('T')
                if channel is 14:
                    self.write_command('Y')
                if channel is 15:
                    self.write_command('U')
                if channel is 16:
                    self.write_command('I')
            # Commands to set toggle to off position
            elif toggle_position == 0:
                if channel is 1:
                    self.write_command('1')
                if channel is 2:
                    self.write_command('2')
                if channel is 3:
                    self.write_command('3')
                if channel is 4:
                    self.write_command('4')
                if channel is 5:
                    self.write_command('5')
                if channel is 6:
                    self.write_command('6')
                if channel is 7:
                    self.write_command('7')
                if channel is 8:
                    self.write_command('8')
                if channel is 9:
                    self.write_command('q')
                if channel is 10:
                    self.write_command('w')
                if channel is 11:
                    self.write_command('e')
                if channel is 12:
                    self.write_command('r')
                if channel is 13:
                    self.write_command('t')
                if channel is 14:
                    self.write_command('y')
                if channel is 15:
                    self.write_command('u')
                if channel is 16:
                    self.write_command('i')
        except Exception as e:
            print("Something went wrong while setting channels: " + str(e))

    # See Cyton SDK for options
    def set_channel_settings(self, channel, enabled=True, gain=24, input_type=0,
                             include_bias=True, use_srb2=True, use_srb1=True):
        try:
            if channel > self.num_channels:
                raise ValueError('Cannot set non-existant channel')
            if self.board_type == 'ganglion':
                raise ValueError('Cannot use with Ganglion')
            ch_array = list("12345678QWERTYUI")
            # defaults
            command = list("x1060110X")
            # Set channel
            command[1] = ch_array[channel - 1]
            # Set power down if needed (default channel enabled)
            if not enabled:
                command[2] = '1'
            # Set gain (default 24)
            if gain == 1:
                command[3] = '0'
            if gain == 2:
                command[3] = '1'
            if gain == 4:
                command[3] = '2'
            if gain == 6:
                command[3] = '3'
            if gain == 8:
                command[3] = '4'
            if gain == 12:
                command[3] = '5'

            # TODO: Implement input type (default normal)

            # Set bias inclusion (default include)
            if not include_bias:
                command[5] = '0'
            # Set srb2 use (default use)
            if not use_srb2:
                command[6] = '0'
            # Set srb1 use (default don't use)
            if use_srb1:
                command[6] = '1'
            command_send = ''.join(command)
            self.write_command(command_send)

            # Make sure to update gain in wifi
            self.gains[channel - 1] = gain
            self.local_wifi_server.set_gains(gains=self.gains)
            self.local_wifi_server.set_parser(
                ParseRaw(gains=self.gains, board_type=self.board_type))

        except ValueError as e:
            print("Something went wrong while setting channel settings: " + str(e))

    def set_sample_rate(self, sample_rate):
        """ Change sample rate """
        try:
            if self.board_type == 'cyton' or self.board_type == 'daisy':
                if sample_rate == 250:
                    self.write_command('~6')
                elif sample_rate == 500:
                    self.write_command('~5')
                elif sample_rate == 1000:
                    self.write_command('~4')
                elif sample_rate == 2000:
                    self.write_command('~3')
                elif sample_rate == 4000:
                    self.write_command('~2')
                elif sample_rate == 8000:
                    self.write_command('~1')
                elif sample_rate == 16000:
                    self.write_command('~0')
                else:
                    print("Sample rate not supported: " + str(sample_rate))
            elif self.board_type == 'ganglion':
                if sample_rate == 200:
                    self.write_command('~7')
                elif sample_rate == 400:
                    self.write_command('~6')
                elif sample_rate == 800:
                    self.write_command('~5')
                elif sample_rate == 1600:
                    self.write_command('~4')
                elif sample_rate == 3200:
                    self.write_command('~3')
                elif sample_rate == 6400:
                    self.write_command('~2')
                elif sample_rate == 12800:
                    self.write_command('~1')
                elif sample_rate == 25600:
                    self.write_command('~0')
                else:
                    print("Sample rate not supported: " + str(sample_rate))
            else:
                print("Board type not supported for setting sample rate")
        except Exception as e:
            print("Something went wrong while setting sample rate: " + str(e))

    def set_accelerometer(self, toggle_position):
        """ Enable / disable accelerometer """
        try:
            if self.board_type == 'ganglion':
                # Commands to set toggle to on position
                if toggle_position == 1:
                    self.write_command('n')
                # Commands to set toggle to off position
                elif toggle_position == 0:
                    self.write_command('N')
            else:
                print("Board type not supported for setting accelerometer")
        except Exception as e:
            print("Something went wrong while setting accelerometer: " + str(e))

    """
    Clean Up (atexit)
    """

    def stop(self):
        print("Stopping streaming...")
        self.streaming = False
        # connection might be already down here
        try:
            if self.impedance:
                print("Stopping with impedance testing")
                self.write_command('Z')
            else:
                self.write_command('s')
        except Exception as e:
            print("Something went wrong while asking the board to stop streaming: " + str(e))
        if self.log:
            logging.warning('sent <s>: stopped streaming')

    def disconnect(self):
        if self.streaming:
            self.stop()

        # should not try to read/write anything after that, will crash

    """
        SETTINGS AND HELPERS
    """

    def warn(self, text):
        if self.log:
            # log how many packets where sent succesfully in between warnings
            if self.log_packet_count:
                logging.info('Data packets received:' +
                             str(self.log_packet_count))
                self.log_packet_count = 0
            logging.warning(text)
        print("Warning: %s" % text)

    def check_connection(self):
        """ Check connection quality in term of lag and number of packets drop.
        Reinit connection if necessary.
        FIXME: parameters given to the board will be lost.
        """
        # stop checking when we're no longer streaming
        if not self.streaming:
            return
        # check number of dropped packets and duration without new packets, deco/reco if too large
        if self.packets_dropped > self.max_packets_to_skip:
            self.warn("Too many packets dropped, attempt to reconnect")
            self.reconnect()
        elif self.timeout > 0 and timeit.default_timer() - self.time_last_packet > self.timeout:
            self.warn("Too long since got new data, attempt to reconnect")
            # if error, attempt to reconect
            self.reconnect()

    def reconnect(self):
        """ In case of poor connection, will shut down and relaunch everything.
        FIXME: parameters given to the board will be lost."""
        self.warn('Reconnecting')
        self.stop()
        self.disconnect()
        self.connect()
        self.init_streaming()


class WiFiShieldHandler(asyncore.dispatcher_with_send):
    def __init__(self, sock, callback=None, high_speed=True,
                 parser=None, daisy=False):
        asyncore.dispatcher_with_send.__init__(self, sock)
        # small recieve buffer so drop samples early if going tooo
        self.socket.setsockopt(socket.SOL_SOCKET,socket.SO_RCVBUF,30000)
        self.callback = callback
        self.daisy = daisy
        self.high_speed = high_speed
        self.last_odd_sample = OpenBCISample()
        self.parser = parser if parser is not None else ParseRaw(
            gains=[24, 24, 24, 24, 24, 24, 24, 24])

    def handle_read(self):
        # 3000 is the max data the WiFi shield is allowed to send over TCP
        data = self.recv(1000)
        if len(data) > 2:
            if self.high_speed:
                packets = int(len(data) / 33)
                raw_data_packets = []
                for i in range(packets):
                    raw_data_packets.append(
                        bytearray(data[i * 33: i * 33 + 33]))
                samples = self.parser.transform_raw_data_packets_to_sample(
                    raw_data_packets=raw_data_packets)

                # if a daisy module is attached, wait to concatenate two samples
                # (main board + daisy) before passing it to callback
                if self.daisy:
                    daisy_samples=[]
                    for sample in samples:
                        # odd sample: daisy sample, save for later
                        if ~sample.sample_number % 2:
                            self.last_odd_sample = sample
                        # even sample: concatenate and send if last sample was the first part,
                        #  otherwise drop the packet
                        elif sample.sample_number - 1 == self.last_odd_sample.sample_number:
                            # the aux data will be the average between the two samples, as the
                            # channel samples themselves have been averaged by the board
                            daisy_samples.append(self.parser.make_daisy_sample_object_wifi(
                                self.last_odd_sample, sample))
                    samples=daisy_samples;
                    
                # now send all the samples to the callback function
                if self.callback is not None:
                    self.callback(samples)
                            
            else:
                try:
                    possible_chunks = data.split('\r\n')
                    if len(possible_chunks) > 1:
                        possible_chunks = possible_chunks[:-1]
                    for possible_chunk in possible_chunks:
                        if len(possible_chunk) > 2:
                            chunk_dict = json.loads(possible_chunk)                            
                            if 'chunk' in chunk_dict:
                                samples=list(chunk_dict['chunk']);
                                if self.callback is not None:
                                    self.callback(samples)
                            else:
                                print("not a sample packet")
                except ValueError as e:
                    print("failed to parse: %s" % data)
                    print(e)
                except BaseException as e:
                    print(e)


class WiFiShieldServer(asyncore.dispatcher):

    def __init__(self, host, port, callback=None, gains=None, high_speed=True, daisy=False):
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind((host, port))
        self.daisy = daisy
        self.listen(5)
        self.callback = None
        self.handler = None
        self.parser = ParseRaw(gains=gains)
        self.high_speed = high_speed

    def handle_accept(self):
        pair = self.accept()
        if pair is not None:
            sock, addr = pair
            print('Incoming connection from %s' % repr(addr))
            self.handler = WiFiShieldHandler(sock, self.callback, high_speed=self.high_speed,
                                             parser=self.parser, daisy=self.daisy)

    def set_callback(self, callback):
        self.callback = callback
        if self.handler is not None:
            self.handler.callback = callback

    def set_daisy(self, daisy):
        self.daisy = daisy
        if self.handler is not None:
            self.handler.daisy = daisy

    def set_gains(self, gains):
        self.parser.set_ads1299_scale_factors(gains)

    def set_parser(self, parser):
        self.parser = parser
        if self.handler is not None:
            self.handler.parser = parser


class ParseRaw(object):
    def __init__(self,
                 board_type='cyton',
                 gains=None,
                 log=False,
                 micro_volts=False,
                 scaled_output=True):
        self.board_type = board_type
        self.gains = gains
        self.log = log
        self.micro_volts = micro_volts
        self.scale_factors = []
        self.scaled_output = scaled_output

        if gains is not None:
            self.scale_factors = self.get_ads1299_scale_factors(self.gains, self.micro_volts)

        self.raw_data_to_sample = RawDataToSample(gains=gains,
                                                  scale=scaled_output,
                                                  scale_factors=self.scale_factors,
                                                  verbose=log)

    def is_stop_byte(self, byte):
        """
        Used to check and see if a byte adheres to the stop byte structure
            of 0xCx where x is the set of numbers from 0-F in hex of 0-15 in decimal.
        :param byte: {int} - The number to test
        :return: {boolean} - True if `byte` follows the correct form
        """
        return (byte & 0xF0) == 0xC0

    def get_ads1299_scale_factors(self, gains, micro_volts=None):
        out = []
        for gain in gains:
            scale_factor = 4.5 / float((pow(2, 23) - 1)) / float(gain)
            if micro_volts is None:
                if self.micro_volts:
                    scale_factor *= 1000000.
            else:
                if micro_volts:
                    scale_factor *= 1000000.

            out.append(scale_factor)
        return out

    def get_channel_data_array(self, raw_data_to_sample):
        """
        :param raw_data_to_sample: RawDataToSample
        :return:
        """
        channel_data = []
        number_of_channels = len(raw_data_to_sample.scale_factors)
        daisy = number_of_channels == 16
        channels_in_packet = 8
        if not daisy:
            channels_in_packet = number_of_channels
        # Channel data arrays are always 8 long

        for i in range(channels_in_packet):
            counts = self.interpret_24_bit_as_int_32(
                raw_data_to_sample.raw_data_packet[
                    (i * 3) +
                    2:(i * 3) +
                    2 + 3
                ]
            )
            channel_data.append(
                raw_data_to_sample.scale_factors[i] *
                counts if raw_data_to_sample.scale else counts
            )

        return channel_data

    def get_data_array_accel(self, raw_data_to_sample):
        accel_data = []
        for i in range(3):
            counts = self.interpret_16_bit_as_int_32(
                raw_data_to_sample.raw_data_packet[
                26 +
                (i * 2): 26 + (i * 2) + 2])
            accel_data.append((0.002 / (pow(2, 4))) *
                              counts if raw_data_to_sample.scale else counts)
        return accel_data

    def get_raw_packet_type(self, stop_byte):
        return stop_byte & 0xF

    def interpret_16_bit_as_int_32(self, two_byte_buffer):
        return struct.unpack('>h', two_byte_buffer)[0]

    def interpret_24_bit_as_int_32(self, three_byte_buffer):
        # 3 byte ints
        unpacked = struct.unpack('3B', three_byte_buffer)

        # 3byte int in 2s compliment
        if unpacked[0] > 127:
            pre_fix = bytes(bytearray.fromhex('FF'))
        else:
            pre_fix = bytes(bytearray.fromhex('00'))

        three_byte_buffer = pre_fix + three_byte_buffer

        # unpack little endian(>) signed integer(i) (makes unpacking platform independent)
        return struct.unpack('>i', three_byte_buffer)[0]

    def parse_packet_standard_accel(self, raw_data_to_sample):
        """
        :param raw_data_to_sample: RawDataToSample
        :return:
        """
        # Check to make sure data is not null.
        if raw_data_to_sample is None:
            raise RuntimeError('Undefined or Null Input')
        if raw_data_to_sample.raw_data_packet is None:
            raise RuntimeError('Undefined or Null Input')

        # Check to make sure the buffer is the right size.
        if len(raw_data_to_sample.raw_data_packet) != 33:
            raise RuntimeError('Invalid Packet Byte Length')

        # Verify the correct start byte.
        if raw_data_to_sample.raw_data_packet[0] != 0xA0:
            raise RuntimeError('Invalid Start Byte')

        sample_object = OpenBCISample()

        sample_object.accel_data = self.get_data_array_accel(raw_data_to_sample)

        sample_object.channels_data = self.get_channel_data_array(raw_data_to_sample)

        sample_object.sample_number = raw_data_to_sample.raw_data_packet[
            1
        ]
        sample_object.start_byte = raw_data_to_sample.raw_data_packet[
            0
        ]
        sample_object.stop_byte = raw_data_to_sample.raw_data_packet[
            32
        ]

        sample_object.valid = True

        now_ms = int(round(time.time() * 1000))

        sample_object.timestamp = now_ms
        sample_object.boardTime = 0

        return sample_object

    def parse_packet_standard_raw_aux(self, raw_data_to_sample):
        pass

    def parse_packet_time_synced_accel(self, raw_data_to_sample):
        pass

    def parse_packet_time_synced_raw_aux(self, raw_data_to_sample):
        pass

    def set_ads1299_scale_factors(self, gains, micro_volts=None):
        self.scale_factors = self.get_ads1299_scale_factors(gains, micro_volts=micro_volts)

    def transform_raw_data_packet_to_sample(self, raw_data):
        """
        Used transform raw data packets into fully qualified packets
        :param raw_data:
        :return:
        """
        try:
            self.raw_data_to_sample.raw_data_packet = raw_data
            packet_type = self.get_raw_packet_type(raw_data[32])
            if packet_type == 0:
                sample = self.parse_packet_standard_accel(self.raw_data_to_sample)
            elif packet_type == 1:
                sample = self.parse_packet_standard_raw_aux(self.raw_data_to_sample)
            elif packet_type == 3 or \
                    packet_type == 4:
                sample = self.parse_packet_time_synced_accel(self.raw_data_to_sample)
            elif packet_type == 5 or \
                    packet_type == 6:
                sample = self.parse_packet_time_synced_raw_aux(self.raw_data_to_sample)
            else:
                sample = OpenBCISample()
                sample.error = 'This module does not support packet type %d' % packet_type
                sample.valid = False

            sample.packet_type = packet_type
        except BaseException as e:
            sample = OpenBCISample()
            if hasattr(e, 'message'):
                sample.error = e.message
            else:
                sample.error = e
            sample.valid = False

        return sample

    def make_daisy_sample_object_wifi(self, lower_sample_object, upper_sample_object):
        """
        /**
        * @description Used to make one sample object from two sample
        *      objects. The sample number of the new daisy sample will be the
        *      upperSampleObject's sample number divded by 2. This allows us
        *      to preserve consecutive sample numbers that flip over at 127
        *      instead of 255 for an 8 channel. The daisySampleObject will
        *      also have one `channelData` array with 16 elements inside it,
        *      with the lowerSampleObject in the lower indices and the
        *      upperSampleObject in the upper set of indices. The auxData from
        *      both channels shall be captured in an object called `auxData`
        *      which contains two arrays referenced by keys `lower` and
        *      `upper` for the `lowerSampleObject` and `upperSampleObject`,
        *      respectively. The timestamps shall be averaged and moved into
        *      an object called `timestamp`. Further, the un-averaged
        *      timestamps from the `lowerSampleObject` and `upperSampleObject`
        *      shall be placed into an object called `_timestamps` which shall
        *      contain two keys `lower` and `upper` which contain the original
        *      timestamps for their respective sampleObjects.
        * @param lowerSampleObject {Object} - Lower 8 channels with odd sample number
        * @param upperSampleObject {Object} - Upper 8 channels with even sample number
        * @returns {Object} - The new merged daisy sample object
        */
        """
        daisy_sample_object = OpenBCISample()

        if lower_sample_object.channels_data is not None:
            daisy_sample_object.channels_data = lower_sample_object.channels_data + \
                upper_sample_object.channels_data

        daisy_sample_object.sample_number = upper_sample_object.sample_number
        daisy_sample_object.id = daisy_sample_object.sample_number

        daisy_sample_object.aux_data = {
            'lower': lower_sample_object.aux_data,
            'upper': upper_sample_object.aux_data
        }

        if lower_sample_object.timestamp:
            daisy_sample_object.timestamp = lower_sample_object.timestamp

        daisy_sample_object.stop_byte = lower_sample_object.stop_byte

        daisy_sample_object._timestamps = {
            'lower': lower_sample_object.timestamp,
            'upper': upper_sample_object.timestamp
        }

        if lower_sample_object.accel_data:
            if lower_sample_object.accel_data[0] > 0 or lower_sample_object.accel_data[1] > 0 or \
                    lower_sample_object.accel_data[2] > 0:
                daisy_sample_object.accel_data = lower_sample_object.accel_data
            else:
                daisy_sample_object.accel_data = upper_sample_object.accel_data

        daisy_sample_object.valid = True

        return daisy_sample_object

    """
    /**
 * @description Used transform raw data packets into fully qualified packets
 * @param o {RawDataToSample} - Used to hold data and configuration settings
 * @return {Array} samples An array of {Sample}
 * @author AJ Keller (@aj-ptw)
 */
function transformRawDataPacketsToSample (o) {
  let samples = [];
  for (let i = 0; i < o.rawDataPackets.length; i++) {
    o.rawDataPacket = o.rawDataPackets[i];
    const sample = transformRawDataPacketToSample(o);
    samples.push(sample);
    if (sample.hasOwnProperty('sampleNumber')) {
      o['lastSampleNumber'] = sample.sampleNumber;
    } else if (!sample.hasOwnProperty('impedanceValue')) {
      o['lastSampleNumber'] = o.rawDataPacket[k.OBCIPacketPositionSampleNumber];
    }
  }
  return samples;
}
    """

    def transform_raw_data_packets_to_sample(self, raw_data_packets):
        samples = []

        for raw_data_packet in raw_data_packets:
            sample = self.transform_raw_data_packet_to_sample(raw_data_packet)
            samples.append(sample)
            self.raw_data_to_sample.last_sample_number = sample.sample_number

        return samples


class RawDataToSample(object):
    """Object encapulsating a parsing object."""

    def __init__(self,
                 accel_data=None,
                 gains=None,
                 last_sample_number=0,
                 raw_data_packets=None,
                 raw_data_packet=None,
                 scale=True,
                 scale_factors=None,
                 time_offset=0,
                 verbose=False):
        """
        RawDataToSample
        :param accel_data: list
            The channel settings array
        :param gains: list
            The gains of each channel, this is used to derive number of channels
        :param last_sample_number: int
        :param raw_data_packets: list
            list of raw_data_packets
        :param raw_data_packet: bytearray
            A single raw data packet
        :param scale: boolean
            Default `true`. A gain of 24 for Cyton will be used and 51 for ganglion by default.
        :param scale_factors: list
            Calculated scale factors
        :param time_offset: int
            For non time stamp use cases i.e. 0xC0 or 0xC1 (default and raw aux)
        :param verbose:
        """
        self.accel_data = accel_data if accel_data is not None else []
        self.gains = gains if gains is not None else []
        self.time_offset = time_offset
        self.last_sample_number = last_sample_number
        self.raw_data_packets = raw_data_packets if raw_data_packets is not None else []
        self.raw_data_packet = raw_data_packet
        self.scale = scale
        self.scale_factors = scale_factors if scale_factors is not None else []
        self.verbose = verbose


class OpenBCISample(object):
    """Object encapulsating a single sample from the OpenBCI board."""

    def __init__(self,
                 aux_data=None,
                 board_time=0,
                 channel_data=None,
                 error=None,
                 imp_data=None,
                 packet_type=0,
                 protocol='wifi',
                 sample_number=0,
                 start_byte=0,
                 stop_byte=0,
                 valid=True,
                 accel_data=None):
        self.aux_data = aux_data if aux_data is not None else []
        self.board_time = board_time
        self.channels_data = channel_data if aux_data is not None else []
        self.error = error
        self.id = sample_number
        self.imp_data = imp_data if aux_data is not None else []
        self.packet_type = packet_type
        self.protocol = protocol
        self.sample_number = sample_number
        self.start_byte = start_byte
        self.stop_byte = stop_byte
        self.timestamp = 0
        self._timestamps = {}
        self.valid = valid
        self.accel_data = accel_data if accel_data is not None else []



#!/usr/bin/python3

#  Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from .utopiaclient import UtopiaClient,Subscribe,PredictedTargetProb,Selection
class Utopia2Output:    
    """Example class for a utopia OUTPUT module.  Connects to the utopia server
    and then either, depending on mode:
        a)  listens for output which exceed it's probability threshold before
    then printing them and using NEWTARGET to indicated the output has taken
    place
    """
    
    def __init__(self,outputPressThreshold=None,outputReleaseThreshold=None,objectID2Action=None):
        self.VERB=0 
        self.outputPressThreshold=outputPressThreshold
        if outputReleaseThreshold is not None :
            self.outputReleaseThreshold=outputPressThreshold*2
        else:
            self.outputReleaseThreshold=outputReleaseThreshold
        # this dictionay contains the functions to execute for the
        # selection IDs we are responsible for. 
        self.objectID2Action=objectID2Action 
        self.client = UtopiaClient()

    def connect(self,host=None,port=None,timeout_ms=30000):
        print('Connecting to utopia on',host,":",port,",",timeout_ms)
        self.client.autoconnect(host,port,timeout_ms)
        self.client.initClockAlign()
        if self.outputPressThreshold is None :
            # selection mode => subscribe only to selection messages
            self.client.sendMessage(Subscribe(self.client.getTimeStamp(),"S"))
            print('In SelectionMode')
        else:
            # perr press/release mode => subscribe to probability output messages
            self.client.sendMessage(Subscribe(self.client.getTimeStamp(),"PS"))
            print('In PerrMode')

    def perrModeOutput(self,msgs):
        """process a perr-message generating appropriate output.  
           To avoid 'key-bounce' we use a press-release semantics, where the output is 'Pressed' 
           so the output is generated when the Perr < outputPressThreshold, then further
           output is inhibited until Perr > outputReleaseThreshold."""
        for msg in msgs:
            if not msg.msgID==PredictedTargetProb.msgID: continue
            #print('OutputnMode:',msg)

            if msg.Perr < self.outputPressThreshold and not self.outputActivated :# low enough error, not already activated
                self.client.sendMessage(Selection(self.client.getTimeStamp(),msg.Yest)) # notify we've made selection
                self.outputActivated=True # update state to output activated
                self.doOutput(msg.objID) # call function to actually make the output happen
            elif msg.Perr > self.outputReleaseThreshold and self.outputActivated : # high-enough error, already activated
                self.outputActivated=False # relase output -> allow a new selection

    def selectionModeOutput(self,msgs):
        """ Process selection message to generate output.  
            Basically generate output if the messages objectID is one of the ones
            we are tasked with generating output for"""
        for msg in msgs:
            if not msg.msgID==Selection.msgID: 
                continue
            print('SelnMode:',msg)
            self.doOutput(msg.objID) # call function to make output happen

    def run(self,timeout_ms=3000):
        """mainloop of utopia-to-output mapping
        runs an infinite loop, waiting for new messages from utopia, filtering out 
        those mesages which contain an output prediction (i.e. PREDICTEDTARGETPROB message)
        and if the output prediction is sufficiently confident forwarding this to the output
        device and sending a NEWTARGET to the recogniser to indicate the output was sent
        """
        if not self.client.isConnected :
            self.connect()
        print("Waiting for messages")
        self.outputActivated=False
        while True:
            newmsgs = self.client.getNewMessages(timeout_ms)
            if not self.outputPressThreshold is None:
                # Perr output mode
                self.perrModeOutput(newmsgs)
            elif self.outputPressThreshold is None:
                # Perr output mode
                self.selectionModeOutput(newmsgs)
                
            print('.',end='')

            
    def doOutput(self,objID):
        """This function is run when objID has sufficiently low error to mean that 
        and output should be generated for this objID. 
        N.B. Override/Replace this function with your specific output method."""
        if self.objectID2Action is None : 
            print("Generated output for Target %d"%(objID))
        else :
            try :
                action = self.objectID2Action[int(objID)]
                action(objID)
            except KeyError:
                pass

''' simple driver for testing '''
if __name__ == "__main__":
    import sys
    host = None
    port = None
    
    if len(sys.argv)>1:
        hostname = sys.argv[1]
        tmp=[hostname.split(":")]
        if len(tmp)>1:
            hostname=tmp[0]
            port    =tmp[1]

    outputthreshold=None
    if len(sys.argv)>2:
        try:
            outputthreshold = float(sys.argv[2])
        except:
            outputthreshold = None
            print('Non-numerical outputthreshold, using selection mode')
            
    u2o = Utopia2Output(outputthreshold)
    u2o.connect(host,port)
    u2o.run()



import os
import signal
from multiprocessing import Process
from time import sleep

def startHubProcess(label):
    from mindaffectBCI.decoder import startUtopiaHub
    hub = Process(target=startUtopiaHub.run, kwargs=dict(label=label), daemon=True)
    hub.start()
    sleep(1)
    return hub

def startAcquisationProcess(label,acquisation,acq_args):
    # start the ganglion acquisation process
    # Using brainflow for the acquisation driver.  
    #  the brainflowargs are kwargs passed to BrainFlowInputParams
    #  so change the board_id and other args to use other boards
    if acquisation == 'none':
        # don't run acq driver here, user will start it manually
        acquisation = None
    elif acquisation == 'fakedata':
        print('Starting fakedata')
        from mindaffectBCI.examples.acquisation import utopia_fakedata
        acq_args=dict(host='localhost', nch=4, fs=200)
        acquisation = Process(target=utopia_fakedata.run, kwargs=acq_args, daemon=True)
        acquisation.start()
    elif acquisation == 'brainflow':
        from mindaffectBCI.examples.acquisation import utopia_brainflow
        if acq_args is None:
            acq_args = dict(board_id=1, serial_port='com3', log=1) # connect to the ganglion
        acquisation = Process(target=utopia_brainflow.run, kwargs=acq_args, daemon=True)
        acquisation.start()

        # give it some time to startup successfully
        sleep(5)
    elif acquisation == 'ganglion': # pyOpenBCI ganglion driver
        from mindaffectBCI.examples.acquisation import utopia_ganglion
        acquisation = Process(target=utopia_ganglion.run, kwargs=acq_args, daemon=True)
        acquisation.start()

    elif acquisation == 'cyton': # pyOpenBCI ganglion driver
        from mindaffectBCI.examples.acquisation import utopia_cyton
        acquisation = Process(target=utopia_cyton.run, kwargs=acq_args, daemon=True)
        acquisation.start()

    elif acquisation == 'javacyton': # java cyton driver
        from mindaffectBCI.examples.acquisation import startJavaCyton
        acquisation = Process(target=startJavaCyton.run, kwargs=acq_args, daemon=True)
        acquisation.start()

    elif acquisation == 'eego': # ANT-neuro EEGO
        from mindaffectBCI.examples.acquisation import utopia_eego
        acquisation = Process(target=utopia_eego.run, kwargs=acq_args, daemon=True)
        acquisation.start()

    elif acquisation == 'lsl': # lsl eeg input stream
        from mindaffectBCI.examples.acquisation import utopia_lsl
        acquisation = Process(target=utopia_lsl.run, kwargs=acq_args, daemon=True)
        acquisation.start()

    else:
        raise ValueError("Unrecognised acquisation driver! {}".format(acquisation))
    
    return acquisation

def startDecoderProcess(label,decoder,decoder_args):
    if decoder == 'decoder' or decoder == 'mindaffectBCI.decoder.decoder':
        from mindaffectBCI.decoder import decoder
        if decoder_args is None:
            decoder_args = dict(calplots=True)
        decoder = Process(target=decoder.run, kwargs=decoder_args, daemon=True)
        decoder.start()
        # allow time for the decoder to startup
        sleep(4)
    elif decoder == 'none':
        decoder = None
    return decoder

def run(label='', acquisation=None, acq_args=None, decoder='decoder', decoder_args=None, presentation='selectionMatrix', presentation_args=None):
    """[summary]

    Args:
        label (str, optional): string label for the saved data file. Defaults to ''.
        acquisation (str, optional): the name of the acquisation driver to use. Defaults to None.
        acq_args (dict, optional): dictionary of optoins to pass to the acquisation driver. Defaults to None.
        decoder (str, optional): the name of the decoder function to use.  Defaults to 'decoder'.
        decoder_args (dict, optional): dictinoary of options to pass to the mindaffectBCI.decoder.run(). Defaults to None.
        presentation (str, optional): the name of the presentation function to use.  Defaults to: 'selectionMatrix'
        presentation_args (dict, optional): dictionary of options to pass to mindaffectBCI.examples.presentation.selectionMatrix.run(). Defaults to None.

    Raises:
        ValueError: invalid options, e.g. unrecognised acq_driver
    """    
    if acquisation is None: 
        acquisation = 'brainflow'

    hub_proc = None
    acquisation_proc = None
    decoder_proc = None
    for retries in range(10):
        #--------------------------- HUB ------------------------------
        # start the utopia-hub process
        if hub_proc is None or not hub_proc.is_alive():
            hub_proc = startHubProcess(label)

        #---------------------------ACQUISATION ------------------------------
        if acquisation_proc is None or not acquisation_proc.is_alive():
            acquisation_proc = startAcquisationProcess(label,acquisation,acq_args)

        #---------------------------DECODER ------------------------------
        # start the decoder process - with default settings for a noise-tag
        if decoder_proc is None or not decoder_proc.is_alive():
            decoder_proc = startDecoderProcess(label, decoder, decoder_args)

        # terminate if all started successfully
        # check all started up and running..
        component_failed=False
        if hub_proc is not None and not hub_proc.is_alive():
            print("Hub didn't start correctly!")
            component_failed=True
        if acquisation_proc is not None and not acquisation_proc.is_alive():
            print("Acq didn't start correctly!")
            component_failed=True
        if decoder_proc is not None and not decoder_proc.is_alive():
            print("Acq didn't start correctly!")
            component_failed=True

        # stop re-starting if all are running fine
        if not component_failed:
            break

    if hub_proc is not None and not hub_proc.is_alive():
        print("Hub didn't start correctly!")
        raise ValueError("Hub didn't start correctly!")
    if acquisation_proc is not None and not acquisation_proc.is_alive():
        print("Acq didn't start correctly!")
        raise ValueError("Acquisation didn't start correctly!")
    if decoder_proc is not None and not decoder_proc.is_alive():
        raise ValueError("Decoder didn't start correctly!")

    #--------------------------- PRESENTATION ------------------------------
    # run the stimulus, with our matrix and default parameters for a noise tag
    #  Make a custom matrix to show:
    if presentation == 'selectionMatrix' or presentation == 'mindaffectBCI.examples.presentation.selectionMatrix':
        if presentation_args is None:
            presentation_args = dict(symbols= [['Hello', 'Good bye'], 
                                               ['Yes',   'No']])
        from mindaffectBCI.examples.presentation import selectionMatrix
        selectionMatrix.run(**presentation_args)
    elif presentation == 'none':
        from mindaffectBCI.decoder.sigViewer import sigViewer
        sigViewer()

    # TODO []: pop-up a monitoring object / dashboard!

    #--------------------------- SHUTDOWN ------------------------------
    # shutdown the background processes
    try: 
        decoder.terminate()
        decoder.join()
    except:
        pass
    try:
        acquisation.terminate()
        acquisation.join()
    except:
        pass
    
    hub_proc.terminate()
    hub_proc.join()
    #print('killing hub')
    #os.kill(hub.pid, signal.SIGTERM)
    #print('exit online_bci')

def parse_args():
    import argparse
    import json
    parser = argparse.ArgumentParser()
    parser.add_argument('--label', type=str, help='user label for the data savefile', default=None)
    parser.add_argument('--config_file', type=str, help='JSON file with default configuration for the on-line BCI', default='online_bci.json')
    parser.add_argument('--acquisation', type=str, help='set the acquisation driver type: one-of: "none","brainflow","fakedata","ganglion","eego"', default=None)
    parser.add_argument('--acq_args', type=json.loads, help='a JSON dictionary of keyword arguments to pass to the acquisation system', default=None)
    parser.add_argument('--decoder', type=str, help='set eeg decoder function to use. one-of: "none", "decoder"', default=None)
    parser.add_argument('--decoder_args', type=json.loads, help='set JSON dictionary of keyword arguments to pass to the decoder. Note: need to doublequote the keywords!', default=None)
    parser.add_argument('--presentation', type=str, help='set stimulus presentation function to use: one-of: "none","selectionMatrix"', default=None)
    parser.add_argument('--presentation_args', type=json.loads, help='set JSON dictionary of keyword arguments to pass to the presentation system', default=None)

    args = parser.parse_args()

    if args.config_file is not None:
        config_file = args.config_file
        # search for the file in py-dir if not in CWD
        import os.path
        if not os.path.isfile(config_file) and os.path.isfile(config_file+'.json'):
                config_file = config_file +'.json'
        if not os.path.isfile(config_file):
            pydir = os.path.dirname(os.path.abspath(__file__))
            if os.path.isfile(os.path.join(pydir,config_file)):
                config_file = os.path.join(pydir, config_file)
            elif os.path.isfile(os.path.join(pydir,config_file+'.json')):
                config_file = os.path.join(pydir, config_file+'.json')

        with open(config_file,'r') as f:
            config = json.load(f)

        # MERGE the config-file parameters with the command-line overrides
        for name in config: # config file parameter
            val = config[name]
            if name in args: # command line override available
                newval = getattr(args, name)
                if newval is None: # ignore not set
                    pass
                elif isinstance(val,dict): # dict, merge with config-file version
                    val.update(newval)
                else: # otherwise just override
                    val = newval
            setattr(args,name,val)
        
        # set label to config file name if label is not set
        if args.label is None:
            # get filename without path or ext
            tmp = os.path.splitext(os.path.basename(config_file))[0]
            setattr(args,'label',tmp)

    return args

# N.B. we need this guard for multiprocessing on Windows!
if __name__ == '__main__':
    args = parse_args()
    run(label=args.label, acquisation=args.acquisation, acq_args=args.acq_args, 
                          decoder=args.decoder, decoder_args=args.decoder_args, 
                          presentation=args.presentation, presentation_args=args.presentation_args)



#!/usr/bin/python3
from bluepy.btle import Peripheral
from bluepy import btle
from sphero_constants import *
import struct
import time
import sys


class sphero_mini():
    def __init__(self, MACAddr, verbosity = 4, user_delegate = None):
        '''
        initialize class instance and then build collect BLE sevices and characteristics.
        Also sends text string to Anti-DOS characteristic to prevent returning to sleep,
        and initializes notifications (which is what the sphero uses to send data back to
        the client).
        '''
        self.verbosity = verbosity # 0 = Silent,
                                   # 1 = Connection/disconnection only
                                   # 2 = Init messages
                                   # 3 = Recieved commands
                                   # 4 = Acknowledgements
        self.sequence = 1
        self.v_batt = None # will be updated with battery voltage when sphero.getBatteryVoltage() is called
        self.firmware_version = [] # will be updated with firware version when sphero.returnMainApplicationVersion() is called

        if self.verbosity > 0:
            print("[INFO] Connecting to", MACAddr)
        self.p = Peripheral(MACAddr, "random") #connect

        if self.verbosity > 1:
            print("[INIT] Initializing")

        # Subscribe to notifications
        self.sphero_delegate = MyDelegate(self, user_delegate) # Pass a reference to this instance when initializing
        self.p.setDelegate(self.sphero_delegate)

        if self.verbosity > 1:
            print("[INIT] Read all characteristics and descriptors")
        # Get characteristics and descriptors
        self.API_V2_characteristic = self.p.getCharacteristics(uuid="00010002-574f-4f20-5370-6865726f2121")[0]
        self.AntiDOS_characteristic = self.p.getCharacteristics(uuid="00020005-574f-4f20-5370-6865726f2121")[0]
        self.DFU_characteristic = self.p.getCharacteristics(uuid="00020002-574f-4f20-5370-6865726f2121")[0]
        self.DFU2_characteristic = self.p.getCharacteristics(uuid="00020004-574f-4f20-5370-6865726f2121")[0]
        self.API_descriptor = self.API_V2_characteristic.getDescriptors(forUUID=0x2902)[0]
        self.DFU_descriptor = self.DFU_characteristic.getDescriptors(forUUID=0x2902)[0]

        # The rest of this sequence was observed during bluetooth sniffing:
        # Unlock code: prevent the sphero mini from going to sleep again after 10 seconds
        if self.verbosity > 1:
            print("[INIT] Writing AntiDOS characteristic unlock code")
        self.AntiDOS_characteristic.write("usetheforce...band".encode(), withResponse=True)

        # Enable DFU notifications:
        if self.verbosity > 1:
            print("[INIT] Configuring DFU descriptor")
        self.DFU_descriptor.write(struct.pack('<bb', 0x01, 0x00), withResponse = True)

        # No idea what this is for. Possibly a device ID of sorts? Read request returns '00 00 09 00 0c 00 02 02':
        if self.verbosity > 1:
            print("[INIT] Reading DFU2 characteristic")
        _ = self.DFU2_characteristic.read()

        # Enable API notifications:
        if self.verbosity > 1:
            print("[INIT] Configuring API dectriptor")
        self.API_descriptor.write(struct.pack('<bb', 0x01, 0x00), withResponse = True)

        self.wake()

        # Finished initializing:
        if self.verbosity > 1:
            print("[INIT] Initialization complete\n")

    def disconnect(self):
        if self.verbosity > 0:
            print("[INFO] Disconnecting")
        
        self.p.disconnect()

    def wake(self):
        '''
        Bring device out of sleep mode (can only be done if device was in sleep, not deep sleep).
        If in deep sleep, the device should be connected to USB power to wake.
        '''
        if self.verbosity > 2:
            print("[SEND {}] Waking".format(self.sequence))
        
        self._send(characteristic=self.API_V2_characteristic,
                   devID=deviceID['powerInfo'],
                   commID=powerCommandIDs["wake"],
                   payload=[]) # empty payload

        #self.getAcknowledgement("Wake")

    def sleep(self, deepSleep=False):
        '''
        Put device to sleep or deep sleep (deep sleep needs USB power connected to wake up)
        '''
        if deepSleep:
            sleepCommID=powerCommandIDs["deepSleep"]
            if self.verbosity > 0:
                print("[INFO] Going into deep sleep. Connect USB power to wake.")
        else:
            sleepCommID=powerCommandIDs["sleep"]
        self._send(characteristic=self.API_V2_characteristic,
                   devID=deviceID['powerInfo'],
                   commID=sleepCommID,
                   payload=[]) #empty payload

    def setLEDColor(self, red = None, green = None, blue = None):
        '''
        Set device LED color based on RGB vales (each can  range between 0 and 0xFF)
        '''
        if self.verbosity > 2:
            print("[SEND {}] Setting main LED colour to [{}, {}, {}]".format(self.sequence, red, green, blue))
        
        self._send(characteristic = self.API_V2_characteristic,
                  devID = deviceID['userIO'], # 0x1a
                  commID = userIOCommandIDs["allLEDs"], # 0x0e
                  payload = [0x00, 0x0e, red, green, blue])

        #self.getAcknowledgement("LED/backlight")

    def setBackLEDIntensity(self, brightness=None):
        '''
        Set device LED backlight intensity based on 0-255 values

        NOTE: this is not the same as aiming - it only turns on the LED
        '''
        if self.verbosity > 2:
            print("[SEND {}] Setting backlight intensity to {}".format(self.sequence, brightness))

        self._send(characteristic = self.API_V2_characteristic,
                  devID = deviceID['userIO'],
                  commID = userIOCommandIDs["allLEDs"],
                  payload = [0x00, 0x01, brightness])

        ##self.getAckknowledgement("LED/backlight")

    def roll(self, speed=None, heading=None):
        '''
        Start to move the Sphero at a given direction and speed.
        heading: integer from 0 - 360 (degrees)
        speed: Integer from 0 - 255
        
        Note: the zero heading should be set at startup with the resetHeading method. Otherwise, it may
        seem that the sphero doesn't honor the heading argument
        '''
        if self.verbosity > 2:
            print("[SEND {}] Rolling with speed {} and heading {}".format(self.sequence, speed, heading))
    
        if abs(speed) > 255:
            print("WARNING: roll speed parameter outside of allowed range (-255 to +255)")

        if speed < 0:
            speed = -1*speed+256 # speed values > 256 in the send packet make the spero go in reverse

        speedH = (speed & 0xFF00) >> 8
        speedL = speed & 0xFF
        headingH = (heading & 0xFF00) >> 8
        headingL = heading & 0xFF
        self._send(characteristic = self.API_V2_characteristic,
                  devID = deviceID['driving'],
                  commID = drivingCommands["driveWithHeading"],
                  payload = [speedL, headingH, headingL, speedH])

        #self.getAcknowledgement("Roll")

    def resetHeading(self):
        '''
        Reset the heading zero angle to the current heading (useful during aiming)
        Note: in order to manually rotate the sphero, you need to call stabilization(False).
        Once the heading has been set, call stabilization(True).
        '''
        if self.verbosity > 2:
            print("[SEND {}] Resetting heading".format(self.sequence))
    
        self._send(characteristic = self.API_V2_characteristic,
                  devID = deviceID['driving'],
                  commID = drivingCommands["resetHeading"],
                  payload = []) #empty payload

        #self.getAcknowledgement("Heading")

    def returnMainApplicationVersion(self):
        '''
        Sends command to return application data in a notification
        '''
        if self.verbosity > 2:
            print("[SEND {}] Requesting firmware version".format(self.sequence))
    
        self._send(self.API_V2_characteristic,
                   devID = deviceID['systemInfo'],
                   commID = SystemInfoCommands['mainApplicationVersion'],
                   payload = []) # empty

        #self.getAcknowledgement("Firmware")

    def getBatteryVoltage(self):
        '''
        Sends command to return battery voltage data in a notification.
        Data printed to console screen by the handleNotifications() method in the MyDelegate class.
        '''
        if self.verbosity > 2:
            print("[SEND {}] Requesting battery voltage".format(self.sequence))
    
        self._send(self.API_V2_characteristic,
                   devID=deviceID['powerInfo'],
                   commID=powerCommandIDs['batteryVoltage'],
                   payload=[]) # empty

        #self.getAcknowledgement("Battery")

    def stabilization(self, stab = True):
        '''
        Sends command to turn on/off the motor stabilization system (required when manually turning/aiming the sphero)
        '''
        if stab == True:
            if self.verbosity > 2:
                    print("[SEND {}] Enabling stabilization".format(self.sequence))
            val = 1
        else:
            if self.verbosity > 2:
                    print("[SEND {}] Disabling stabilization".format(self.sequence))
            val = 0
        self._send(self.API_V2_characteristic,
                   devID=deviceID['driving'],
                   commID=drivingCommands['stabilization'],
                   payload=[val])

        #self.getAcknowledgement("Stabilization")

    def wait(self, delay):
        '''
        This is a non-blocking delay command. It is similar to time.sleep(), except it allows asynchronous 
        notification handling to still be performed.
        '''
        start = time.time()
        while(1):
            self.p.waitForNotifications(0.001)
            if time.time() - start > delay:
                break

    def _send(self, characteristic=None, devID=None, commID=None, payload=[]):
        '''
        A generic "send" method, which will be used by other methods to send a command ID, payload and
        appropriate checksum to a specified device ID. Mainly useful because payloads are optional,
        and can be of varying length, to convert packets to binary, and calculate and send the
        checksum. For internal use only.

        Packet structure has the following format (in order):

        - Start byte: always 0x8D
        - Flags byte: indicate response required, etc
        - Virtual device ID: see sphero_constants.py
        - Command ID: see sphero_constants.py
        - Sequence number: Seems to be arbitrary. I suspect it is used to match commands to response packets (in which the number is echoed).
        - Payload: Could be varying number of bytes (incl. none), depending on the command
        - Checksum: See below for calculation
        - End byte: always 0xD8

        '''
        sendBytes = [sendPacketConstants["StartOfPacket"],
                    sum([flags["resetsInactivityTimeout"], flags["requestsResponse"]]),
                    devID,
                    commID,
                    self.sequence] + payload # concatenate payload list

        self.sequence += 1 # Increment sequence number, ensures we can identify response packets are for this command
        if self.sequence > 255:
            self.sequence = 0

        # Compute and append checksum and add EOP byte:
        # From Sphero docs: "The [checksum is the] modulo 256 sum of all the bytes
        #                   from the device ID through the end of the data payload,
        #                   bit inverted (1's complement)"
        # For the sphero mini, the flag bits must be included too:
        checksum = 0
        for num in sendBytes[1:]:
            checksum = (checksum + num) & 0xFF # bitwise "and to get modulo 256 sum of appropriate bytes
        checksum = 0xff - checksum # bitwise 'not' to invert checksum bits
        sendBytes += [checksum, sendPacketConstants["EndOfPacket"]] # concatenate

        # Convert numbers to bytes
        output = b"".join([x.to_bytes(1, byteorder='big') for x in sendBytes])

        #send to specified characteristic:
        characteristic.write(output, withResponse = True)

    def getAcknowledgement(self, ack):
        #wait up to 10 secs for correct acknowledgement to come in, including sequence number!
        start = time.time()
        while(1):
            self.p.waitForNotifications(1)
            if self.sphero_delegate.notification_seq == self.sequence-1: # use one less than sequence, because _send function increments it for next send. 
                if self.verbosity > 3:
                    print("[RESP {}] {}".format(self.sequence-1, self.sphero_delegate.notification_ack))
                self.sphero_delegate.clear_notification()
                break
            elif self.sphero_delegate.notification_seq >= 0:
                print("Unexpected ACK. Expected: {}/{}, received: {}/{}".format(
                    ack, self.sequence, self.sphero_delegate.notification_ack.split()[0],
                    self.sphero_delegate.notification_seq),
                    file=sys.stderr)
            if time.time() > start + 1:
                print("Timeout waiting for acknowledgement: {}/{}".format(ack, self.sequence), file=sys.stderr)
                break

# =======================================================================
# The following functions are experimental:
# =======================================================================

    def configureCollisionDetection(self,
                                     xThreshold = 50, 
                                     yThreshold = 50, 
                                     xSpeed = 50, 
                                     ySpeed = 50, 
                                     deadTime = 50, # in 10 millisecond increments
                                     method = 0x01, # Must be 0x01        
                                     callback = None):
        '''
        Appears to function the same as other Sphero models, however speed settings seem to have no effect. 
        NOTE: Setting to zero seems to cause bluetooth errors with the Sphero Mini/bluepy library - set to 
        255 to make it effectively disabled.

        deadTime disables future collisions for a short period of time to avoid repeat triggering by the same
        event. Set in 10ms increments. So if deadTime = 50, that means the delay will be 500ms, or half a second.
        
        From Sphero docs:
        
            xThreshold/yThreshold: An 8-bit settable threshold for the X (left/right) and Y (front/back) axes 
            of Sphero.

            xSpeed/ySpeed: An 8-bit settable speed value for the X and Y axes. This setting is ranged by the 
            speed, then added to xThreshold, yThreshold to generate the final threshold value.
        '''
        
        if self.verbosity > 2:
            print("[SEND {}] Configuring collision detection".format(self.sequence))
    
        self._send(self.API_V2_characteristic,
                   devID=deviceID['sensor'],
                   commID=sensorCommands['configureCollision'],
                   payload=[method, xThreshold, xSpeed, yThreshold, ySpeed, deadTime])

        self.collision_detection_callback = callback

        #self.getAcknowledgement("Collision")

    def configureSensorStream(self): # Use default values
        '''
        Send command to configure sensor stream using default values as found during bluetooth 
        sniffing of the Sphero Edu app.

        Must be called after calling configureSensorMask()
        '''
        bitfield1 = 0b00000000 # Unknown function - needs experimenting
        bitfield2 = 0b00000000 # Unknown function - needs experimenting
        bitfield3 = 0b00000000 # Unknown function - needs experimenting
        bitfield4 = 0b00000000 # Unknown function - needs experimenting

        if self.verbosity > 2:
            print("[SEND {}] Configuring sensor stream".format(self.sequence))
    
        self._send(self.API_V2_characteristic,
                   devID=deviceID['sensor'],
                   commID=sensorCommands['configureSensorStream'],
                   payload=[bitfield1, bitfield1, bitfield1, bitfield1])

        #self.getAcknowledgement("Sensor")

    def configureSensorMask(self,
                            sample_rate_divisor = 0x25, # Must be > 0
                            packet_count = 0,
                            IMU_pitch = False,
                            IMU_roll = False,
                            IMU_yaw = False,
                            IMU_acc_x = False,
                            IMU_acc_y = False,
                            IMU_acc_z = False,
                            IMU_gyro_x = False,
                            IMU_gyro_y = False,
                            IMU_gyro_z = False):

        '''
        Send command to configure sensor mask using default values as found during bluetooth 
        sniffing of the Sphero Edu app. From experimentation, it seems that these are he functions of each:
        
        Sampling_rate_divisor. Slow data EG: Set to 0x32 to the divide data rate by 50. Setting below 25 (0x19) causes 
                bluetooth errors        
        
        Packet_count: Select the number of packets to transmit before ending the stream. Set to zero to stream infinitely
        
        All IMU bool parameters: Toggle transmission of that value on or off (e.g. set IMU_acc_x = True to include the 
                X-axis accelerometer readings in the sensor stream)
        '''

        # Construct bitfields based on function parameters:
        IMU_bitfield1 = (IMU_pitch<<2) + (IMU_roll<<1) + IMU_yaw
        IMU_bitfield2 = ((IMU_acc_y<<7) + (IMU_acc_z<<6) + (IMU_acc_x<<5) + \
                         (IMU_gyro_y<<4) + (IMU_gyro_x<<2) + (IMU_gyro_z<<2))
        
        if self.verbosity > 2:
            print("[SEND {}] Configuring sensor mask".format(self.sequence))
    
        self._send(self.API_V2_characteristic,
                   devID=deviceID['sensor'],
                   commID=sensorCommands['sensorMask'],
                   payload=[0x00,               # Unknown param - altering it seems to slow data rate. Possibly averages multiple readings?
                            sample_rate_divisor,       
                            packet_count,       # Packet count: select the number of packets to stop streaming after (zero = infinite)
                            0b00,               # Unknown param: seems to be another accelerometer bitfield? Z-acc, Y-acc
                            IMU_bitfield1,
                            IMU_bitfield2,
                            0b00])              # reserved, Position?, Position?, velocity?, velocity?, Y-gyro, timer, reserved

        #self.getAcknowledgement("Mask")

        '''
        Since the sensor values arrive as unlabelled lists in the order that they appear in the bitfields above, we need 
        to create a list of sensors that have been configured.Once we have this list, then in the default_delegate class, 
        we can get sensor values as attributes of the sphero_mini class.
        e.g. print(sphero.IMU_yaw) # displays the current yaw angle
        '''

        # Initialize dictionary with sensor names as keys and their bool values (set by the user) as values:
        availableSensors = {"IMU_pitch" : IMU_pitch,
                            "IMU_roll" : IMU_roll,
                            "IMU_yaw" : IMU_yaw,
                            "IMU_acc_y" : IMU_acc_y,
                            "IMU_acc_z" : IMU_acc_z,
                            "IMU_acc_x" : IMU_acc_x,
                            "IMU_gyro_y" : IMU_gyro_y,
                            "IMU_gyro_x" : IMU_gyro_x,
                            "IMU_gyro_z" : IMU_gyro_z}
        
        # Create list of of only sensors that have been "activated" (set as true in the method arguments):
        self.configured_sensors = [name for name in availableSensors if availableSensors[name] == True]

    def sensor1(self): # Use default values
        '''
        Unknown function. Observed in bluetooth sniffing. 
        '''
        self._send(self.API_V2_characteristic,
                   devID=deviceID['sensor'],
                   commID=sensorCommands['sensor1'],
                   payload=[0x01])

        #self.getAcknowledgement("Sensor1")

    def sensor2(self): # Use default values
        '''
        Unknown function. Observed in bluetooth sniffing. 
        '''
        self._send(self.API_V2_characteristic,
                   devID=deviceID['sensor'],
                   commID=sensorCommands['sensor2'],
                   payload=[0x00])

        #self.getAcknowledgement("Sensor2")

# =======================================================================

class MyDelegate(btle.DefaultDelegate):

    '''
    This class handles notifications (both responses and asynchronous notifications).
    
    Usage of this class is described in the Bluepy documentation
    
    '''

    def __init__(self, sphero_class, user_delegate):
        self.sphero_class = sphero_class # for saving sensor values as attributes of sphero class instance
        self.user_delegate = user_delegate # to directly notify users of callbacks
        btle.DefaultDelegate.__init__(self)
        self.clear_notification()
        self.notificationPacket = []

    def clear_notification(self):
        self.notification_ack = "DEFAULT ACK"
        self.notification_seq = -1

    def bits_to_num(self, bits):
        '''
        This helper function decodes bytes from sensor packets into single precision floats. Encoding follows the
        the IEEE-754 standard.
        '''
        num = int(bits, 2).to_bytes(len(bits) // 8, byteorder='little')
        num = struct.unpack('f', num)[0]
        return num

    def handleNotification(self, cHandle, data):
        '''
        This method acts as an interrupt service routine. When a notification comes in, this
        method is invoked, with the variable 'cHandle' being the handle of the characteristic that
        sent the notification, and 'data' being the payload (sent one byte at a time, so the packet
        needs to be reconstructed)  

        The method keeps appending bytes to the payload packet byte list until end-of-packet byte is
        encountered. Note that this is an issue, because 0xD8 could be sent as part of the payload of,
        say, the battery voltage notification. In future, a more sophisticated method will be required.
        '''
        # Allow the user to intercept and process data first..
        if self.user_delegate != None:
            if self.user_delegate.handleNotification(cHandle, data):
                return

        for data_byte in data: # parse each byte separately (sometimes they arrive simultaneously)

            self.notificationPacket.append(data_byte) # Add new byte to packet list

            # If end of packet (need to find a better way to segment the packets):
            if data_byte == sendPacketConstants['EndOfPacket']:
                # Once full the packet has arrived, parse it:
                # Packet structure is similar to the outgoing send packets (see docstring in sphero_mini._send())
                
                # Attempt to unpack. Might fail if packet is too badly corrupted
                try:
                    start, flags_bits, devid, commcode, seq, *notification_payload, chsum, end = self.notificationPacket
                except ValueError:
                    print("Warning: notification packet unparseable", self.notificationPacket, file=sys.stderr)
                    self.notificationPacket = [] # Discard this packet
                    return # exit

                # Compute and append checksum and add EOP byte:
                # From Sphero docs: "The [checksum is the] modulo 256 sum of all the bytes
                #                   from the device ID through the end of the data payload,
                #                   bit inverted (1's complement)"
                # For the sphero mini, the flag bits must be included too:
                checksum_bytes = [flags_bits, devid, commcode, seq] + notification_payload
                checksum = 0 # init
                for num in checksum_bytes:
                    checksum = (checksum + num) & 0xFF # bitwise "and to get modulo 256 sum of appropriate bytes
                checksum = 0xff - checksum # bitwise 'not' to invert checksum bits
                if checksum != chsum: # check computed checksum against that recieved in the packet
                    print("Warning: notification packet checksum failed", self.notificationPacket, file=sys.stderr)
                    self.notificationPacket = [] # Discard this packet
                    return # exit

                # Check if response packet:
                if flags_bits & flags['isResponse']: # it is a response

                    # Use device ID and command code to determine which command is being acknowledged:
                    if devid == deviceID['powerInfo'] and commcode == powerCommandIDs['wake']:
                        self.notification_ack = "Wake acknowledged" # Acknowledgement after wake command
                        
                    elif devid == deviceID['driving'] and commcode == drivingCommands['driveWithHeading']:
                        self.notification_ack = "Roll command acknowledged"

                    elif devid == deviceID['driving'] and commcode == drivingCommands['stabilization']:
                        self.notification_ack = "Stabilization command acknowledged"

                    elif devid == deviceID['userIO'] and commcode == userIOCommandIDs['allLEDs']:
                        self.notification_ack = "LED/backlight color command acknowledged"

                    elif devid == deviceID['driving'] and commcode == drivingCommands["resetHeading"]:
                        self.notification_ack = "Heading reset command acknowledged"

                    elif devid == deviceID['sensor'] and commcode == sensorCommands["configureCollision"]:
                        self.notification_ack = "Collision detection configuration acknowledged"

                    elif devid == deviceID['sensor'] and commcode == sensorCommands["configureSensorStream"]:
                        self.notification_ack = "Sensor stream configuration acknowledged"

                    elif devid == deviceID['sensor'] and commcode == sensorCommands["sensorMask"]:
                        self.notification_ack = "Mask configuration acknowledged"

                    elif devid == deviceID['sensor'] and commcode == sensorCommands["sensor1"]:
                        self.notification_ack = "Sensor1 acknowledged"

                    elif devid == deviceID['sensor'] and commcode == sensorCommands["sensor2"]:
                        self.notification_ack = "Sensor2 acknowledged"

                    elif devid == deviceID['powerInfo'] and commcode == powerCommandIDs['batteryVoltage']:
                        V_batt = notification_payload[2] + notification_payload[1]*256 + notification_payload[0]*65536
                        V_batt /= 100 # Notification gives V_batt in 10mV increments. Divide by 100 to get to volts.
                        self.notification_ack = "Battery voltage:" + str(V_batt) + "v"
                        self.sphero_class.v_batt = V_batt

                    elif devid == deviceID['systemInfo'] and commcode == SystemInfoCommands['mainApplicationVersion']:
                        version = '.'.join(str(x) for x in notification_payload)
                        self.notification_ack = "Firmware version: " + version
                        self.sphero_class.firmware_version = notification_payload
                                                
                    else:
                        self.notification_ack = "Unknown acknowledgement" #print(self.notificationPacket)
                        print(self.notificationPacket, "===================> Unknown ack packet")

                    self.notification_seq = seq

                else: # Not a response packet - therefore, asynchronous notification (e.g. collision detection, etc):
                    
                    # Collision detection:
                    if devid == deviceID['sensor'] and commcode == sensorCommands['collisionDetectedAsync']:
                        # The first four bytes are data that is still un-parsed. the remaining unsaved bytes are always zeros
                        _, _, _, _, _, _, axis, _, Y_mag, _, X_mag, *_ = notification_payload
                        if axis == 1: 
                            dir = "Left/right"
                        else:
                            dir = 'Forward/back'
                        print("Collision detected:")
                        print("\tAxis:", dir)
                        print("\tX_mag:", X_mag)
                        print("\tY_mag:", Y_mag)

                        #if self.sphero_class.collision_detection_callback is not None:
                         #   self.notificationPacket = [] # need to clear packet, in case new notification comes in during callback
                          #  self.sphero_class.collision_detection_callback()

                    # Sensor response:
                    elif devid == deviceID['sensor'] and commcode == sensorCommands['sensorResponse']:
                        # Convert to binary, pad bytes with leading zeros:
                        val = ''
                        for byte in notification_payload:
                            val += format(int(bin(byte)[2:], 2), '#010b')[2:]
                        
                        # Break into 32-bit chunks
                        nums = []
                        while(len(val) > 0):
                            num, val = val[:32], val[32:] # Slice off first 16 bits
                            nums.append(num)
                        
                        # convert from raw bits to float:
                        nums = [self.bits_to_num(num) for num in nums]

                        # Set sensor values as class attributes:
                        for name, value in zip(self.sphero_class.configured_sensors, nums):
                            setattr(self.sphero_class, name, value)
                        
                    # Unrecognized packet structure:
                    else:
                        self.notification_ack = "Unknown asynchronous notification" #print(self.notificationPacket)
                        print(self.notificationPacket, "===================> Unknown async packet")
                        
                self.notificationPacket = [] # Start new payload after this byte



import numpy as np
from mindaffectBCI.decoder.utils import window_axis, idOutliers, zero_outliers
#@function
def updateSummaryStatistics(X, Y, stimTimes=None, 
                            Cxx=None, Cxy=None, Cyy=None, 
                            badEpThresh=4, halflife_samp=1, cxxp=True, cyyp=True, tau=None, offset=0, center=True):
    '''
    Compute updated summary statistics (Cxx, Cxy, Cyy) for new data in X with event-info Y
      [Cxx, Cxy, Cyy, state]=updateSummaryStatistics(X, Y, stimTime_samp, Cxx, Cxy, Cyy, state, halflife_samp, badEpThresh)
    Inputs:
      X = (nTrl, nEp, tau, d) [d x tau x nEpoch x nTrl] raw response for the current stimulus event
               d=#electrodes tau=#response-samples  nEpoch=#stimulus events to process
         OR
          (nTrl, nSamp, d) [d x nSamp x nTrl ]
      Y = (nTrl, nEp, nY, nE) [nE x nY x nEpoch x nTrl] Indicator for which events occured for which outputs
         OR
           (nTrl, nSamp, nY, nE) [nE x nY x nSamp x nTrl]
               nE=#event-types  nY=#possible-outputs  nEpoch=#stimulus events to process
      stimTimes_samp = (nTrl, nEp) [nEpoch x nTrl] sample times for start each epoch.  Used to detect
               overlapping responses
      tau = int - the length of the impulse response in samples (X.shape[1] or Cxy.shape[1] if given)
      offset:int = relative shift of Y w.r.t. X
      Cxx = [d x d] current data covariance
      Cxy = (nY, nE, tau, d) [d x tau x nE x nY ] current per output ERPs
      Cyy = (nY, nE, tau, nE, tau) [tau x nE x tau x nE x nY ] current response covariance for each output
      badEpThresh = float, threshold for removing bad-data before fitting the summary statistics
      center:bool - do we center the X data before computing the summary statistics? (True)
      halflife_samp = float, forgetting factor for the updates
               Note: alpha = exp(log(.5)./(half-life)), half-life = log(.5)/log(alpha)

    
    Outputs:
      Cxx = [d x d] updated data covariance
      Cxy = (nY, nE, tau, d) [d x tau x nE x nY] updated per output ERPs
      Cyy = (nY, nE, tau, nE, tau) [tau x nE x tau x nE x nY] updated response covariance for each output
    Examples:
      # Supervised CCA
      Y = (nEpoch/nSamp, nY, nE) [nE x nY x nEpoch] indicator for each event-type and output of it's type in each epoch
      X = (nEpoch/nSamp, tau, d) [d x tau x nEpoch] sliced pre-processed raw data into per-stimulus event responses
      stimTimes = [nEpoch] sample numbers of the stimulus events
                # OR None if X, Y are samples
      [Cxx, Cxy, Cyy] = updateSummaryStatistics(X, Y, 3*np.arange(X.shape[0]));
      [J, w, r]=multipleCCA(Cxx, Cxy, Cyy)
    '''
    if X is None:
        return Cxx, Cxy, Cyy
    if tau is None:
        if Cxy is not None:
            tau = Cxy.shape[-2]
        elif X.shape[-2] < 100: # assume X is already sliced so we can infer the tau..
            print("Warning: guessing tau from shape of X")
            tau = X.shape[-2]
        else:
            raise ValueError("tau not set and Cxy is None!")
    tau = int(tau) # ensure tau is integer
    wght = 1

    X, Y = zero_outliers(X, Y, badEpThresh)
    
    # update the cross covariance XY
    Cxy = updateCxy(Cxy, X, Y, stimTimes, tau, wght, offset, center)
    # Update Cxx
    if (cxxp):
        Cxx = updateCxx(Cxx, X, stimTimes, None, wght, center)

    # ensure Cyy has the right size if not entry-per-model
    if (cyyp):
        # TODO [] : support overlapping info between update calls
        Cyy = updateCyy(Cyy, Y, stimTimes, tau, wght)
        
    return Cxx, Cxy, Cyy

#@function
def updateCxx(Cxx, X, stimTimes=None, tau=None, wght=1, center=False, unitnorm=True):
    '''
    Cxx = [d x d] current data covariance
    X = (nTrl, nEp, tau, d) raw response for the current stimulus event
            d=#electrodes tau=#response-samples  nEpoch=#stimulus events to process
        OR
        (nTrl, nSamp, d) 
    stimTimes_samp = (nTrl, nEp) sample times for start each epoch.  Used to detect
    wght=[1x1]
    center:bool - compute centered covariance? (True)
    '''
    # ensure 3d
    if X.ndim == 2:
        X = X[np.newaxis, ...]

    # use X as is, computing purely spatial cov
    if tau is None:
        # N.B. reshape is expensive with window_axis
        if X.ndim == 3:
            XX = np.einsum("TEd,TEe->de", X, X)
            N = X.shape[0]
        elif X.ndim == 4:
            XX = np.einsum("TEtd,TEte->de", X, X)
            N = X.shape[0]*X.shape[1]

        if center:
            sX = np.sum(X.reshape((-1,X.shape[-1])),axis=0) # feature mean (d,)
            XX = XX - np.einsum("i,j->ij",sX,sX) *X.shape[-1]/X.size
        

    else: # compute time-shifted covariance
        Xe = window_axis(X, winsz=tau, axis=-2) # (nTrl, nSamp, tau, d)
        # N.B. reshape is expensive if use window axis
        XX = np.einsum("TEtd,TEue->tdue",Xe, Xe)

        if center:
            raise NotImplementedError("Centering with shifts!")

    if unitnorm:
        XX = XX / N

    Cxx = wght*Cxx + XX if Cxx is not None else XX
    return Cxx

#@function
def updateCxy(Cxy, X, Y, stimTimes=None, tau=None, wght=1, offset=0, center=False, verb=0, unitnorm=True):
    '''
    X = (nTrl, nEp, tau, d) raw response for the current stimulus event
            d=#electrodes tau=#response-samples  nEpoch=#stimulus events to process
        OR
        (nTrl, nSamp, d) raw response at sample rate
    Y = (nTrl, nEp, nY, nE) Indicator for which events occured for which outputs
            nE=#event-types  nY=#possible-outputs  nEpoch=#stimulus events to process
        OR 
        (nTrl, nSamp, nY, nE) event indicator at sample rate
    stimTimes_samp = (nTrl, nEp) sample times for start each epoch.  Used to detect
            overlapping responses
    Cxy = (nY, nE, tau, d) current per output ERPs
    '''
    if tau is None: # estimate the tau
        tau = Cxy.shape[-2]
    if Y.ndim == 3:
        Y = Y[np.newaxis, :, :, :] # add missing trial dim
    if offset<-tau+1:
        raise NotImplementedError("Cant offset backwards by more than window size")
    if offset>0:
        raise NotImplementedError("Cant offset by positive amounts")
    if verb > 1: print("tau={}".format(tau))
    if stimTimes is None:
        # X, Y are at sample rate, slice X every sample
        Xe = window_axis(X, winsz=tau, axis=-2) # (nTrl, nSamp, tau, d)
        Ye = Y[:, -offset:Xe.shape[-3]-offset, :, :] # shrink Y to fit w.r.t. window
    else:
        Xe = X
        Ye = Y
    if Xe.ndim == 3:
        Xe = Xe[np.newaxis, :, :, :]

    if verb > 1: print("Xe={}\nYe={}".format(Xe.shape, Ye.shape))
    XY = np.einsum("TEye, TEtd->yetd", Ye, Xe)

    if center:
        if verb > 1: print("center")
        # N.B. (X-mu*1.T)@Y.T = X@Y.T - (mu*1.T)@Y.T = X@Y.T - mu*1.T@Y.T = X@Y.T - mu*\sum(Y)  
        muX = np.mean(X.reshape((-1,X.shape[-1])),axis=0) # mean X (d,)
        muY = np.sum(Ye,(0,1)) # summed Y (nY, nE)
        muXY= np.einsum("ye,d->yed",muY,muX) #(nY,nE,d)
        XY = XY - muXY[...,np.newaxis,:] #(nY,nE,tau,d)

    if unitnorm:
        #print('Cxy - unitnorm!')
        #if verb>0 : 
        #print("tau={}".format(tau))
        #XY = XY / np.sqrt(tau)
        pass
    
    Cxy = wght*Cxy + XY if Cxy is not None else XY
    return Cxy

# @function
def updateCyy(Cyy, Y, stimTime=None, tau=None, wght=1, zeropadded=True, unitnorm=True):
    '''
    Compute the Cyy tensors given new data
    Inputs:
      Cyy -- (nY, nE, tau, nE, tau) old Cyy info
      Y -- (nTrl, nEp/nSamp, nY, nE) the new stim info to be added
               Indicator for which events occured for which outputs
               nE=#event-types  nY=#possible-outputs  nEpoch=#stimulus events to process
      tau  -- number of samples in the stimulus response
      stimTime -- [] the times (in samples) of the epochs in Y
                  None : if Y is already at sample rate
      zeropadded - bool - flag variable length Y are padded with 0s
      wght -- weighting for the new vs. old data
      unitnorm(bool) : flag if we normalize the Cyy with number epochs
    Outputs:
      Cyy -- (nY, nE, tau, nE, tau)
    '''
    if tau is None: # estimate the tau
        tau=Cyy.shape[-1]
    if Cyy is None:
        Cyy = np.zeros((Y.shape[-2], Y.shape[-1], tau, Y.shape[-1], tau))
    if Y.ndim == 3:  # ensure is 4-d
        Y = Y[np.newaxis, :, :, :] # (nTrl, nEp/nSamp, nY, nE) [nE x nY x nEpoch/nSamp x nTrl]

    if stimTime is None: # fast-path, already at sample rate
        if np.issubdtype(Y.dtype, np.floating): # all at once
            #print("Y={}".format(Y.shape))
            Ys = window_axis(Y, winsz=tau, axis=-3) # window of length tau (nTrl, nSamp, tau, nY, nE) [ nE x nY x nSamp x tau x nTrl ]
            #print("Ys={}".format(Ys.shape))
            MM = np.einsum("TStye, TSuyf->yetfu", Ys, Ys) # compute cross-covariance (nY, nE, tau, nE, tau) [ nE x tau x nE x tau x nY ]
        else: # trial at a time + convert to float
            MM = np.zeros(Cyy.shape)
            # different slice time for every trial, up-sample per-trial
            for trli in range(Y.shape[0]):  # loop over trials
                Yi = Y[trli, :, :, :]
                if not np.issubdtype(Y.dtype, np.floating):
                    Yi = np.array(Yi, np.float)
                Yi = window_axis(Yi, winsz=tau, axis=-3)
                MM = MM+np.einsum("Etye, Euyf->yetfu", Yi, Yi)

    else: # upsample and accumulate
        if stimTime.ndim == 1:  # ensure is 2d
            stimTime = stimTime[np.newaxis, :]

        MM = np.zeros(Cyy.shape)
        # different slice time for every trial, up-sample per-trial
        for trli in range(Y.shape[0]):  # loop over trials
            # up-sample to sample rate
            if trli < stimTime.shape[0]:
                nEp    = np.flatnonzero(stimTime[trli, :] != 0)[-1]
                sampIdx = np.array(stimTime[trli, :nEp] - stimTime[trli, 0], dtype=int)
            Yi = np.zeros((sampIdx[-1]+1+tau, )+Y.shape[-2:])
            Yi[sampIdx, :, :] = Y[trli, :nEp, :, :] # (nSamp, nY, nE)
            # compute the outer product for the covariance
            Yi = window_axis(Yi, winsz=tau, axis=-3) # (nEp/nSamp, tau, nY, nE)
            #print("Yi={}".format(Yi.shape))
            MM = MM + np.einsum("Etye, Euyf->yetfu", Yi, Yi) # (nY, nE, tau, nE, tau)
    
    if unitnorm:
        # normalize so the resulting constraint on the estimated signal is that it have
        # average unit norm
        MM = MM / (Y.shape[0]*Y.shape[1]) # / nTrl*nEp

    # accumulate into the running total
    Cyy = wght*Cyy + MM if Cyy is not None else Cyy
    return Cyy


def autocov(X, tau):
    '''
    Compute the cross-auto-correlation of X, i.e. spatial-temporal covariance
    Inputs:
      X -- (nTrl, nSamp, d) the data to have autocov computed
               nSamp = temporal dim, d = spatial dim
      tau  -- number of samples in the autocov
    Outputs:
      Ctdtd -- (tau, d, tau, d)
    '''
    if X.ndim == 2: # ensure 3-d
        X = X[np.newaxis, :, :]
    if X.ndim > 4: # ensure 3-d
        X = X.reshape((np.prod(X.shape[:-2]),)+X.shape[-2:])

    if np.issubdtype(X.dtype, np.float): # all at once
        Xs = window_axis(X, winsz=tau, axis=-2) # window of length tau (nTrl, nSamp-tau, tau, d)
        Ctdtd=np.einsum("Tstd, Tsue -> tdue", Xs, Xs, optimize='optimal') # compute cross-covariance (tau, d, tau, d)

    else: # trial at a time + convert to float
        Ctdtd = np.zeros((tau, X.shape[-1], tau, X.shape[-1]))
        # different slice time for every trial, up-sample per-trial
        for ti in range(X.shape[0]):  # loop over trials            
            Xi = X[ti, :, :,]
            if not np.issubdtype(X.dtype, np.float):
                Xi = np.array(Xi, np.float)
            Xi = window_axis(Xi, winsz=tau, axis=-2)
            Ctdtd = Ctdtd + np.einsum("Etd, Eue->tdue", Xi, Xi)
            
    return Ctdtd

def cov(X):
    '''
    Compute the spatial covariance
    Input:
     X = (...,d) data, with features in the last dimension
    Output:
     Cxx = (d,d)
    '''
    # make X 2d
    X2d = np.reshape(X, (np.prod(X.shape[:-1]), X.shape[-1]))
    Cxx  = np.einsum("Td, Te->de", X2d, X2d)
    return Cxx

def crossautocov(X, Y, tau, offset=0):
    '''
    Compute the cross-auto-correlation between 2 datasets, X,Y, i.e. the double spatial-temporal covariance
    Inputs:
      X -- (nTrl, nSamp, dx) the data to have autocov computed
               nSamp = temporal dim, dx = spatial dim x
      Y -- (nTrl, nSamp, dy) the data to have autocov computed
               nSamp = temporal dim, dy = spatial dim y
      tau  -- :int number of samples in the cov
             OR
              [int,int] time lags for x and y
    Outputs:
      Ctdtd -- (taux, dx, tauy, dy)
    '''
    if X.ndim == 2: # ensure 3-d
        X = X[np.newaxis, :, :]
    if X.ndim > 4: # ensure 3-d
        X = X.reshape((np.prod(X.shape[:-2]),)+X.shape[-2:])
    if Y.ndim == 2: # ensure 3-d
        Y = Y[np.newaxis, :, :]
    if Y.ndim > 4: # ensure 3-d
        Y = Y.reshape((np.prod(Y.shape[:-2]),)+Y.shape[-2:])
    if not hasattr(tau, '__iter__'): # scalar tau, same for x and y
        tau = (tau, tau)
    if offset!=0:
        raise NotImplementedError("Offsets not supported yet...")
        
    #TODO[]: optimize for memory usage?
    if X.shape[0] < 100 and np.issubdtype(X.dtype, np.float) and np.issubdtype(Y.dtype, np.float): # all at once
        Xs = window_axis(X, winsz=tau[0], axis=-2) # window of length tau (nTrl, nSamp-tau, tau, d)
        Ys = window_axis(Y, winsz=tau[1], axis=-2) # window of length tau (nTrl, nSamp-tau, tau, d)
        
        # reshape back to same shape if diff window sizes
        if tau[0] < tau[1]:
            Xs = Xs[..., :Ys.shape[-3], :, :]
        else:
            Ys = Ys[..., -offset:Xs.shape[-3]-offset, :, :]

        # compute cross-covariance (tau, d, tau, d)
        Ctdtd = np.einsum("Tstd, Tsue -> tdue", Xs, Ys, optimize='optimal')
        
    else: # trial at a time + convert to float
        Ctdtd = np.zeros((tau[0], X.shape[-1], tau[1], Y.shape[-1]))
        # different slice time for every trial, up-sample per-trial
        for ti in range(X.shape[0]):  # loop over trials
            
            Xi = X[ti, :, :]
            if not np.issubdtype(X.dtype, np.float32) and not np.issubdtype(X.dtype,np.float64):
                Xi = np.array(Xi, np.float)
            Xi = window_axis(Xi, winsz=tau[0], axis=-2)
            
            Yi = Y[ti, :, :]
            if not np.issubdtype(Y.dtype, np.float32) and not np.issubdtype(Y.dtype,np.float64):
                Yi = np.array(Yi, np.float)
            Yi = window_axis(Yi, winsz=tau[1], axis=-2)
            
            # reshape back to same shape if diff window-sizes
            if tau[0] < tau[1]:
                Xi = Xi[..., :Yi.shape[-3], :, :]
            else:
                Yi = Yi[..., offset:Xi.shape[-3]+offset, :, :]

            # compute cross-covariance (tau, d, tau, d)
            Ctdtd = Ctdtd + np.einsum("Etd, Eue -> tdue", Xi, Yi)
        
    return Ctdtd

def plot_summary_statistics(Cxx, Cxy, Cyy, evtlabs=None, times=None, ch_names=None, fs=None):
    '''
    Cxx =(d, d) updated data covariance
    Cxy = (nY, nE, tau, d)  updated per output ERPs
    Cyy = (nY, nE, tau, nE, tau) updated response covariance for each output
    '''
    import matplotlib.pyplot as plt
    if times is None:
        times = np.arange(Cxy.shape[-2])
        if fs is not None:
            times = times / fs
    if ch_names is None:
        ch_names = ["{}".format(i) for i in range(Cxy.shape[-1])]
    if evtlabs is None:
        evtlabs = ["{}".format(i) for i in range(Cxy.shape[-3])]

    plt.clf()
    # Cxx
    plt.subplot(311);
    plt.imshow(Cxx, origin='lower', extent=[0, Cxx.shape[0], 0, Cxx.shape[1]])
    # TODO []: use the ch_names to add lables to the  axes
    plt.title('Cxx')

    # Cxy
    if Cxy.ndim > 3:
        if Cxy.shape[0] > 1:
            print("Warning: only the 1st set ERPs is plotted")
        Cxy = Cxy[0, ...]
    nevt = Cxy.shape[-3]
    for ei in range(nevt):
        if ei==0:
            ax = plt.subplot(3,nevt,nevt+ei+1) # N.B. subplot indexs from 1!
            plt.xlabel('time (s)')
            plt.ylabel('space')
        else: # no axis on other sub-plots
            plt.subplot(3, nevt, nevt+ei+1, sharey=ax, sharex=ax)
            plt.tick_params(labelbottom=False, labelleft=False)
        plt.imshow(Cxy[ei, :, :].T, aspect='auto', origin='lower', extent=(times[0], times[-1], 0, Cxy.shape[-1]))
        # TODO []: use the ch_names to add lables to the  axes
        plt.title('{}'.format(evtlabs[ei]))
    # only last one has colorbar
    plt.colorbar()

    # Cyy
    if Cyy.ndim > 4:
        if Cyy.shape[0] > 1:
            print("Warning: only the 1st set ERPs is plotted")
        Cyy = Cyy[0, ...]
    Cyy2d = np.reshape(Cyy, (Cyy.shape[0]*Cyy.shape[1], Cyy.shape[2]*Cyy.shape[3]))
    plt.subplot(313)
    plt.imshow(Cyy2d, origin='lower', extent=[0, Cyy2d.shape[0], 0, Cyy2d.shape[1]])
    plt.title('Cyy')

def plot_erp(erp, evtlabs=None, times=None, fs=None, ch_names=None, axis=-1, plottype='plot', offset=0, ylim=None):
    '''
    Make a multi-plot of the event ERPs (as stored in erp)
    erp = (nE, tau, d) current per output ERPs
    '''
    if erp.ndim > 3:
        if erp.shape[0]>1 :
            print("Warning: only the 1st set ERPs is plotted")
        erp = erp[0, ...]
    elif erp.ndim == 2:
        erp = erp[np.newaxis,...]
    icoords = evtlabs if not evtlabs is None else list(range(erp.shape[-3]))
    if times is None:
        times = list(range(offset,erp.shape[-2]+offset))
        if fs is not None:
            times = [ t/fs for t in times ]
    jcoords = times if not times is None else list(range(offset,erp.shape[-2]+offset))
    kcoords = ch_names if not ch_names is None else list(range(erp.shape[-1]))
    if axis<0:
        axis = erp.ndim+axis
    import matplotlib.pyplot as plt

    print("erp={}".format(erp.shape))
    
    # plot per channel
    ncols = int(np.ceil(np.sqrt(erp.shape[axis])))
    nrows = int(np.ceil(erp.shape[axis]/ncols))
    # make the bottom left axis to share its limits..
    axploti = ncols*(nrows-1)
    ax = plt.subplot(nrows,ncols,axploti+1)
    #fig, plts = plt.subplots(nrows, ncols, sharex='all', sharey='all', squeeze=False)
    for ci in range(erp.shape[axis]):
        # make the axis
        if ci==axploti: # common axis plot
            pl = ax
        else: # normal plot
            pl = plt.subplot(nrows,ncols,ci+1,sharex=ax, sharey=ax) # share limits
            plt.tick_params(labelbottom=False,labelleft=False) # no labels

        # get the slice for the data to plot,  and it's coords
        if axis == 0:
            A = erp[ci, :, :] # (time,ch)
            pltcoords = icoords
            xcoords = jcoords
            ycoords = kcoords
        elif axis == 1:
            A = erp[:, ci, :] # (evtcoords,ch)
            xcoords = icoords
            pltcoords = jcoords
            ycoords = kcoords            
        elif axis == 2:
            A = erp[:, :, ci] # (evtlabs,time)
            xcoords = icoords
            ycoords = jcoords
            pltcoords = kcoords

        # make the plot of the desired type
        if plottype == 'plot':
            for li in range(A.shape[0]):
                pl.plot(ycoords, A[li, :], '.-', label=xcoords[li], markersize=1, linewidth=1)
                if ylim: plt.ylim(ylim)
        elif plottype == 'plott':
            for li in range(A.shape[1]):
                pl.plot(xcoords, A[:, li], '.-', label=ycoords[li], markersize=1, linewidth=1)
                if ylim: plt.ylim(ylim)
        elif plottype == 'imshow':
            # TODO[] : add the labels to the rows
            pl.imshow(A, aspect='auto')#,extent=[times[0],0,times[-1],erp.shape[-3]])
            
        pl.title.set_text("{}".format(pltcoords[ci]))
    # legend only in the last plot
    pl.legend()


def plot_factoredmodel(A, R, evtlabs=None, times=None, ch_names=None, ch_pos=None, fs=None, spatial_filter_type="Filter"):
    '''
    Make a multi-plot of a factored model
    A = (k,d)
    R = (k,e,tau)
    '''
    if A.ndim > 2:
        if A.shape[0]>1 :
            print("Warning: only the 1st set ERPs is plotted")
        A = A[0, ...]
    if A.ndim < 2:
        A = A[np.newaxis, :]
    if R.ndim > 3:
        if R.shape[0]>1 :
            print("Warning: only the 1st set ERPs is plotted")
        R = R[0, ...]
    if R.ndim < 3:
        R = R[np.newaxis, :]

    if times is None:
        times = np.arange(R.shape[-1])
        if fs is not None:
            times = times / fs
    if ch_names is None:
        ch_names = np.arange(A.shape[-1])
    elif ch_pos is None and len(ch_names) > 0:
        # try to load position info from capfile
        try: 
            print("trying to get pos from cap file!")
            from mindaffectBCI.decoder.readCapInf import getPosInfo
            cnames, xy, xyz, iseeg =getPosInfo(ch_names)
            if all(iseeg):
                ch_pos = xy
        except:
            pass
    if evtlabs is None:
        evtlabs = np.arange(R.shape[-2])
        
    import matplotlib.pyplot as plt

    print("A={} R={}".format(A.shape, R.shape))
    
    # plot per component
    ncols = 2 #int(np.ceil(np.sqrt(A.shape[0])))
    nrows = A.shape[0] #int(np.ceil(A.shape[0]/ncols))
    # start at the bottom to share the axis
    for ci in range(A.shape[0]):
        # make the axis
        subploti=(nrows-1-ci)*ncols
        if ci==0: # common axis plot
            axR = plt.subplot(nrows, ncols, subploti+2) # share limits
            axA = plt.subplot(nrows, ncols, subploti+1) # share limits
            axA.set_xlabel("Space")
            axR.set_xlabel("time (s)")
            axA.grid(True)
            axR.grid(True)
            pA = axA
            pR = axR
        else: # normal plot
            pR = plt.subplot(nrows, ncols, subploti+2, sharex=axR, sharey=axR) 
            pA = plt.subplot(nrows, ncols, subploti+1, sharex=axA, sharey=axA) 
            plt.tick_params(labelbottom=False,labelleft=False) # no labels
            pA.grid(True)
            pR.grid(True)

        # make the spatial plot
        sign = np.sign(A[ci,np.argmax(np.abs(A[ci,:]))]) # normalize directions
        if not ch_pos is None:
            cRng= np.max(np.abs(A.reshape((-1))))
            levels = np.linspace(-cRng,cRng,20)
            tt=pA.tricontourf(ch_pos[:,0],ch_pos[:,1],A[ci,:]*sign,levels=levels,cmap='Spectral')
            pA.plot(ch_pos[:,0],ch_pos[:,1],'ko',markersize=5)
            plt.colorbar(tt)
        else:
            pA.plot(ch_names,A[ci,:]*sign,'.-')
        pA.title.set_text("Spatial {} #{}".format(spatial_filter_type,ci))
        # make the temporal plot, with labels, N.B. use loop so can set each lines label
        for e in range(R.shape[-2]):
            pR.plot(times,R[ci,e,:]*sign,'.-',label=evtlabs[e])
        pR.title.set_text("Impulse Response #{}".format(ci))

    # legend in the temporal plot
    axR.legend()
    # rotate the ticks to be readable?
    if ch_pos is None:
        # TODO[]: get and only set for the tick locations!
        #axA.set_xticklabels(ch_names,rotation=65)
        pass


def testSlicedvsContinuous():
    import numpy as np
    from utils import testSignal, sliceData, sliceY
    irf=(0,0,0,0,0,1,0,0,0,0)
    offset=0; # o->lag-by-5, -5=>no-lag, -9->lead-by-4
    X,Y,st,A,B = testSignal(nTrl=1,nSamp=500,d=1,nE=1,nY=1,isi=10,tau=10,offset=offset,irf=irf,noise2signal=0)
    #plt.subplot(311);plt.plot(X[0,:,0],label='X');plt.plot(Y[0,:,0,0],label='Y');plt.title("offset={}, irf={}".format(offset,irf));plt.legend()
    print("A={}\nB={}".format(A, B))
    print("X{}={}".format(X.shape, X[:30, np.argmax(np.abs(A))]))
    print("Y{}={}".format(Y.shape, Y[0, :30, 0]))
    Y_true = Y[..., 0:1, :]

    # slice then compute
    Xe=sliceData(X, stimTimes, tau=tau)
    Ye=sliceY(Y, stimTimes)
    Ye_true = Ye[..., 0:1, :]
    print('Xe{}={}\nYe{}={}\nst{}={}'.format(Xe.shape, Xe[:5, :, 0], Ye.shape, Ye[:5, 0, 0], stimTimes.shape, stimTimes[:5]))
    Cxx, Cxy, Cyy = updateSummaryStatistics(Xe, Ye_true, stimTimes, tau=Xe.shape[-2])
    print("Cxx={} Cxy={} Cyy={}".format(Cxx.shape, Cxy.shape, Cyy.shape))

    plot_summary_statistics(Cxx, Cxy, Cyy)

    plot_Cxy(Cxy)
    
    dCxx = np.max((np.abs(Cxx/np.trace(Cxx)-Cxx1/np.trace(Cxx1))/(np.abs(Cxx/np.trace(Cxx))+1e-5)).ravel())
    dCxy = np.max((np.abs(Cxy-Cxy1)/(np.abs(Cxy)+1e-5)).ravel())
    dCyy = np.max((np.abs(Cyy-Cyy1)/(np.abs(Cyy)+1e-5)).ravel())
    print("dCxx= {} dCxy={} dCyy={}".format(dCxx, dCxy, dCyy))

    # timing tests
    from timeit import timeit
    dur = timeit(lambda: updateSummaryStatistics(Xe, Ye_true, stimTimes, tau=10), number=1000, globals=globals())
    print("Sliced={}s".format(dur/1000))

    def slicenuss(X, Y, stimTimes, tau):
        Xe = sliceData(X, stimTimes, tau)
        Ye = sliceY(Y, stimTimes)
        Cxx, Cxy, Cyy = updateSummaryStatistics(Xe, Ye, stimTimes, tau=tau)
    dur = timeit(lambda: slicenuss(X, Y_true, stimTimes, tau=10), number=1000, globals=globals())
    print("Slice+USS={}s".format(dur/1000))

    dur = timeit(lambda: updateSummaryStatistics(X, Y_true, None, tau=10), number=1000, globals=globals())
    print("Raw={}s".format(dur/1000))


def testComputationMethods():
    import numpy as np
    from utils import testSignal, sliceData, sliceY
    irf=(0,0,0,0,0,1,0,0,0,0)
    offset=0; # o->lag-by-5, -5=>no-lag, -9->lead-by-4
    X,Y,st,A,B = testSignal(nTrl=1,nSamp=500,d=1,nE=1,nY=1,isi=10,tau=10,offset=offset,irf=irf,noise2signal=0)
    #plt.subplot(311);plt.plot(X[0,:,0],label='X');plt.plot(Y[0,:,0,0],label='Y');plt.title("offset={}, irf={}".format(offset,irf));plt.legend()
    print("A={}\nB={}".format(A, B))
    print("X{}={}".format(X.shape, X[:30, np.argmax(np.abs(A))]))
    print("Y{}={}".format(Y.shape, Y[0, :30, 0]))
    Y_true = Y[..., 0:1, :]

    # other measures....
    from updateSummaryStatistics import updateSummaryStatistics, autocov, cov, crossautocov, updateCyy, updateCxx, updateCxy
    Cxdtxdt = autocov(X, tau=18) # (tau[0],dx,tau[1],dx)
    Cxdtydt = crossautocov(X, Y[:, :, 0:1], tau=18) # (tau[0],dx,tau[1],dy)
    Cydxdt  = crossautocov(Y[:, :, 0:1], X, tau=[1,18]) # (tauy,dy,taux,dx)
    Cydtydt = np.stack([crossautocov(Y[:, :, yi:yi+1], Y[:, :, yi:yi+1], tau=18) for yi in range(Y.shape[-1])], 0) # (nY,tau,d,tau,d)
    Cxxdd   = cov(X)

    # compare
    Cxx=updateCxx(None,X,None)
    Cxy=updateCxy(None,X,Y[:,:,0:1,np.newaxis],None,tau=18)#(1,dy,tau,dx)
    Cyy=updateCyy(None,Y[:,:,:,np.newaxis],None,tau=18) #(nY,yd,tau,yd,tau)

    print('Cxx-Cxxdd={}'.format(np.max(np.abs(Cxxdd-Cxx).ravel())))
    print('Cxy-Cydxdt={}'.format(np.max(np.abs(Cxy-Cydxdt).ravel())))
    print('Cyy-Cydtydt={}'.format(np.max(np.abs(Cyy.ravel()-Cydtydt.ravel()))))

    # test centering
    cX = X - np.mean(X,(0,1))
    Cxx = updateCxx(None,X,None,center=True)
    cCxx = updateCxx(None,cX,None,center=False)
    print("Cxx(center) - cCxx={}".format(np.max(np.abs(Cxx-cCxx).ravel())))

    Cxy = updateCxy(None,X,Y,None,tau=3,center=True)
    cCxy= updateCxy(None,cX,Y,None,tau=3,center=False)
    print("Cxy(center) - cCxy={}".format(np.max(np.abs(Cxy-cCxy).ravel())))
    
    
def testCases():
    import numpy as np
    from utils import testSignal, sliceData, sliceY
    from updateSummaryStatistics import updateSummaryStatistics, plot_summary_statistics
    import matplotlib.pyplot as plt

    irf=(0,0,0,0,0,0,0,0,0,1)
    offset=0; # X->lag-by-10
    X,Y,st,A,B = testSignal(nTrl=1,nSamp=500,d=1,nE=1,nY=1,isi=10,tau=10,offset=offset,irf=irf,noise2signal=0)
    print("A={}\nB={}".format(A, B))
    print("X{}={}".format(X.shape, X[:30, np.argmax(np.abs(A))]))
    print("Y{}={}".format(Y.shape, Y[0, :30, 0]))
    
    tau=10; uss_offset=0
    Cxx, Cxy, Cyy = updateSummaryStatistics(X, Y, None, tau=tau, offset=uss_offset)
    print("Cxx={} Cxy={} Cyy={}".format(Cxx.shape, Cxy.shape, Cyy.shape))
    plt.figure(1);plot_summary_statistics(Cxx,Cxy,Cyy); plt.show()
    
    # leading X
    irf=(1,0,0,0,0,0,0,0,0,0)
    offset=-9; # X leads by 9
    X,Y,st,A,B = testSignal(nTrl=1,nSamp=5000,d=1,nE=1,nY=1,isi=10,tau=10,offset=offset,irf=irf,noise2signal=0)
    plt.figure(0);plt.clf();plt.plot(X[0,:,0],label='X');plt.plot(Y[0,:,0,0],label='Y');plt.title("offset={}, irf={}".format(offset,irf));plt.legend()

    # no-shift in analysis window
    tau=10; uss_offset=0
    Cxx, Cxy, Cyy = updateSummaryStatistics(X, Y, None, tau=tau, offset=uss_offset)
    print("Cxx={} Cxy={} Cyy={}".format(Cxx.shape, Cxy.shape, Cyy.shape))
    plt.figure(1);plt.clf();plot_summary_statistics(Cxx,Cxy,Cyy);plt.show()

    # shifted analysis window
    tau=10; uss_offset=-9
    Cxx, Cxy, Cyy = updateSummaryStatistics(X, Y, None, tau=tau, offset=uss_offset)
    print("Cxx={} Cxy={} Cyy={}".format(Cxx.shape, Cxy.shape, Cyy.shape))
    plt.figure(2);plt.clf();plot_summary_statistics(Cxx,Cxy,Cyy);plt.show()

    plt.figure(3);plt.clf();plot_erp(Cxy);plt.show()

    
if __name__=="__main__":
    testCases()



import numpy as np
from mindaffectBCI.decoder.analyse_datasets import debug_test_dataset, analyse_dataset, analyse_datasets
from mindaffectBCI.decoder.offline.load_mindaffectBCI  import load_mindaffectBCI
from mindaffectBCI.decoder.timestamp_check import timestampPlot
import matplotlib.pyplot as plt

savefile = '~/Downloads/mindaffectBCI_200701_1617[1].txt'

savefile = '~/Downloads/mindaffectBCI_200701_1624[1].txt'

savefile = '~/Downloads/mindaffectBCI_200701_2025_khash.txt'

savefile = '~/Desktop/mark/mindaffectBCI*ganglion*1411*.txt'
savefile = '~/Desktop/mark/mindaffectBCI*.txt'

import glob
import os
files = glob.glob(os.path.expanduser(savefile)); 
#os.path.join(os.path.dirname(os.path.abspath(__file__)),fileregexp)) # * means all if need specific format then *.csv
savefile = max(files, key=os.path.getctime)

# load
X, Y, coords = load_mindaffectBCI(savefile, stopband=((45,65),(5.5,25,'bandpass')), order=6, ftype='butter', fs_out=100)
# output is: X=eeg, Y=stimulus, coords=meta-info about dimensions of X and Y
print("EEG: X({}){} @{}Hz".format([c['name'] for c in coords],X.shape,coords[1]['fs']))
print("STIMULUS: Y({}){}".format([c['name'] for c in coords[:1]]+['output'],Y.shape))

plt.close('all')
# train *only* on 1st 10 trials
debug_test_dataset(X, Y, coords, cv=[(slice(10),slice(10,None))], tau_ms=450, evtlabs=('re','fe'), rank=1, model='cca')


# do a time-stamp check.
timestampPlot(savefile)

# # check the electrode qualities computation
# ppfn= butterfilt_and_downsample(order=6, stopband='butter_stopband((0, 5), (25, -1))_fs200.pk', fs_out=100)
# X=X.reshape((-1,X.shape[-1]))
# ppfn.fit(X[:10,:],fs=200)
# Xp =  ppfn.transform(X)
# plt.clf();plt.subplot(211);plt.plot(X);plt.subplot(212);plt.plot(Xp);

# sigq=testElectrodeQualities(X,fs=200)
# plt.clf();plt.plot(sigq)





import numpy as np
from mindaffectBCI.decoder.normalizeOutputScores import normalizeOutputScores
from mindaffectBCI.decoder.zscore2Ptgt_softmax import zscore2Ptgt_softmax
#@function
def decodingSupervised(Fy, softmaxscale=3.5, badFyThresh=4,
                       centFy=True, detrendFy=True, 
                       nEpochCorrection=30,
                       minDecisLen=0, maxDecisLen=0,
                       bwdAccumulate=False,
                       marginalizemodels=True, 
                       nocontrolamplitude=None,
                       priorsigma=(-1,120),
                       tiebreaking_noise=1e-8):
    '''
    true-target estimator and error-probility estimator for each trial
    Inputs:
      Fy   - (nModel,nTrl,nEp,nY) [#Y x #Epoch x #Trials x nModels]
      softmaxscale - the scale length to pass to zscore2Ptgt_softmax.py
      badFyThresh - threshold for detction of bad Fy entry, in std-dev
      centFy   - bool, do we center Fy before computing  *important*  (true)
      nEpochCorrection - int, number of epochs to use a base for correction of number epochs in the sum.
               such that: zscore = Fy / ( sigma * ( 1+1/(nEpoch/nEpochErrorCorrection)) ),
               basically, nEpoch < nEpochCorrection have highly increased Perr, so unlikely to be selected
      minDecisLen - int, number of epochs to use as base for distribution of time-based decision points.
                         i.e. decisions at, [1,2,4,...2^n]*exptDistDecis
                    OR: minDecisLen<0 => decision point every abs(minDeicsLen) epochs
      bwdAccumulate - [bool], accumulate data backwards from last epoch gathered
      maxDecisLen   - maximum number of epochs for a decision
      priorsigma = (sigma,N) prior estimate of sigma2 and number pseudo-points (10,30)
    Outputs:
      Yest - (nTrl,nDecis) [ nDecis x nTrl ] the most likely / minimum error output for each decision point
      Perr - (nTrl,nDecis) [ nDecix x nTrl ] the probability that this selection is an ERROR for each decision point
      Ptgt - (nTrl,nDecis,nY) [ nY x nDecis x nTrl ] the probability each target is the true target for each decision point

    Copyright (c) MindAffect B.V. 2018
    '''
    if Fy is None:
        return -1, 1, None, None, None
    
    # get the info on which outputs are zero in each trial
    validTgt = np.any(Fy != 0, axis=-2) # valid if non-zero for any epoch..# (nModel,nTrl,nY)  
    # normalize the raw scores for each model to have nice distribution
    ssFy,varsFy,decisIdx,nEp,nY=normalizeOutputScores(Fy, validTgt=validTgt,
                                               badFyThresh=badFyThresh, 
                                               centFy=centFy, detrendFy=detrendFy,
                                               nEpochCorrection=nEpochCorrection,
                                               minDecisLen=minDecisLen, maxDecisLen=maxDecisLen,
                                               bwdAccumulate=bwdAccumulate,
                                               priorsigma=priorsigma)

    if nocontrolamplitude is not None:
      raise NotImplementedError('no-control signal not yet implemented correctly')
      # # add a no-control pseudo-output to simulate not looking
      # #Median rather than mean?
      # mussFy=np.sum(ssFy,0)/np.maximum(2,np.sum(validTgt,0,keepdims=True)) # mean score for each trial/decisPt [ 1 x nDecis x nTrl x nMdl ]
      # # add to the scores & validTgt info
      # ssFy=np.concatenate((ssFy,nocontrolamplitude+mussFy),0)
      # validtgtTrl=np.concatenate((validtgtTrl,np.ones([1,size(validtgtTrl,2),size(validtgtTrl,3),size(validtgtTrl,4)])))

    # compute the target probabilities over output for each model+trial
    # use the softmax approach to get Ptgt for all outputs
    Ptgt = zscore2Ptgt_softmax(ssFy,
                               softmaxscale,
                               validTgt=validTgt,
                               marginalizemodels=marginalizemodels) # (nM,nTrl,nDecis,nY) [ nY x nDecis x nTrl]
    # extract the predicted output and it's probability of being the target
    Ptgt2d = Ptgt.reshape((np.prod(Ptgt.shape[:-1]), Ptgt.shape[-1])) # make 2d-copy
    # add tie-breaking noise
    if tiebreaking_noise > 0:
        Ptgt2d = Ptgt2d + np.random.standard_normal(Ptgt2d.shape)*tiebreaking_noise

    Yestidx = np.argmax(Ptgt2d, -1) # max over outputs
    Ptgt_max = Ptgt2d[np.arange(Ptgt2d.shape[0]), Yestidx] # value at max, indexing trick to find..
    Yestidx = Yestidx.reshape(Ptgt.shape[:-1]) # -> (nM,nTrl,nY)
    Ptgt_max = Ptgt_max.reshape(Ptgt.shape[:-1])# -> (nM,nTrl,nY)

    # pick a model for this maxY
    decisMdl = Yestidx if ssFy.ndim>3 else 1
    decisEp = 1   
    Yest = Yestidx
    # p(maxz != tgt ) = 1 - p(maxz==tgt)
    Perr = 1 - Ptgt_max
    return Yest, Perr, Ptgt, decisMdl, decisEp


def decodingSupervised_streamed(Fy,softmaxscale=3,nocontrolamplitude=None,badFyThresh=6,
                                centFy=True,nEpochCorrection=15,
                                minDecisLen=0,maxDecisLen=0,
                                bwdAccumulate=False, filtLen=15,
                                marginalizemodels=False):
    '''
    true-target estimator and error-probility estimator for each trial
    Inputs:
      Fy   - (nModel,nTrl,nEp,nY) [#Y x #Epoch x #Trials x nModels]
      softmaxscale - the scale length to pass to zscore2Ptgt_softmax.py
      badFyThresh - threshold for detction of bad Fy entry, in std-dev
      centFy   - bool, do we center Fy before computing  *important*  (true)
      nEpochCorrection - int, number of epochs to use a base for correction of number epochs in the sum.
               such that: zscore = Fy / ( sigma * ( 1+1/(nEpoch/nEpochErrorCorrection)) ),
               basically, nEpoch < nEpochCorrection have highly increased Perr, so unlikely to be selected
      minDecisLen - int, number of epochs to use as base for distribution of time-based decision points.
                         i.e. decisions at, [1,2,4,...2^n]*exptDistDecis
                    OR: minDecisLen<0 => decision point every abs(minDeicsLen) epochs
      bwdAccumulate - [bool], accumulate data backwards from last epoch gathered
      maxDecisLen   - maximum number of epochs for a decision
    Outputs:
      Yest - (nTrl,nDecis) [ nDecis x nTrl ] the most likely / minimum error output for each decision point
      Perr - (nTrl,nDecis) [ nDecix x nTrl ] the probability that this selection is an ERROR for each decision point
      Ptgt - (nTrl,nDecis,nY) [ nY x nDecis x nTrl ] the probability each target is the true target for each decision point

    Copyright (c) MindAffect B.V. 2018
    '''
    if Fy is None:
        return -1, 1, None, None, None
    
    # get the info on which outputs are zero in each trial
    validTgt = np.any(Fy != 0, axis=-2) # valid if non-zero for any epoch..# (nModel,nTrl,nY) [ nY x nTrl x nModel ]   
    # normalize the raw scores for each model to have nice distribution
    stdFy,varsFy,N,nEp,nY=normalizeOutputScores_streamed(Fy, validTgt=validTgt,
                                                         badFyThresh=badFyThresh, centFy=centFy,
                                                         minDecisLen=minDecisLen, filtLen=filtLen)



    # compute the target probabilities over output for each block+model+trial
    # use the softmax approach to get Ptgt for all outputs
    Ptgt = zscore2Ptgt_softmax(ssFy,
                               softmaxscale,
                               validTgt=validTgt,
                               marginalizemodels=False) # (nM,nTrl,nDecis,nY) [ nY x nDecis x nTrl]
    # extract the predicted output and it's probability of being the target
    Ptgt2d = Ptgt.reshape((np.prod(Ptgt.shape[:-1]), Ptgt.shape[-1])) # make 2d-copy
    Yestidx = np.argmax(Ptgt2d, -1) # max over outputs
    Ptgt_max = Ptgt2d[np.arange(Ptgt2d.shape[0]), Yestidx] # value at max, indexing trick to find..
    Yestidx = Yestidx.reshape(Ptgt.shape[:-1]) # -> (nM,nTrl,nY)
    Ptgt_max = Ptgt_max.reshape(Ptgt.shape[:-1])# -> (nM,nTrl,nY)

    # pick a model for this maxY
    decisMdl = Yestidx if ssFy.ndim>3 else 1
    decisEp = 1   
    Yest = Yestidx
    # p(maxz != tgt ) = 1 - p(maxz==tgt)
    Perr = 1 - Ptgt_max
    return Yest, Perr, Ptgt, decisMdl, decisEp



#@function
def testcase():
    from normalizeOutputScores import mktestFy
    Fy,nEp=mktestFy(startupNoisefrac=0,trlenfrac=0,sigstr=.4)
    from decodingSupervised import decodingSupervised
    print("Fy={}".format(Fy.shape))
    Yest,Perr,Ptgt,decismdl,decisEp=decodingSupervised(Fy)
    print("Fy={}".format(Fy[:,[0],:,:].shape))
    Yest,maxp,ssFy,decismdl,decisEp=decodingSupervised(Fy[:,[0],:,:])
    np.mean(Yest == 0,1) 
    from decodingCurveSupervised import decodingCurveSupervised
    decodingCurveSupervised(Fy);

if __name__=="__main__":
    testcase()




import numpy as np
from mindaffectBCI.decoder.utils import window_axis
#@function
def scoreOutput(Fe, Ye, dedup0=None, R=None, offset=0, outputscore='ip'):
    '''
    score each output given information on which stim-sequences corrospend to which inputs
    Inputs:
      Fe  = (nM,nTrl,nSamp,nE) similarity score for each event type for each stimulus
      Ye  = (nTrl,nSamp,nY,nE) Indicator for which events occured for which outputs
               nE=#event-types  nY=#possible-outputs  nEpoch=#stimulus events to process
      R = (nM,nfilt,nE,tau) FWD-model (impulse response) for each of the event types, used to correct the 
            scores for correlated responses.
      dedup0 = bool - remove duplicate copies of output O (used when cross validating calibration data)
    Outputs:
      Fy  = (nM,nTrl,nEp,nY) similarity score for each input epoch for each output
    
    Copyright (c) MindAffect B.V. 2018
    '''
    if Fe.size == 0:
        Fy = np.zeros((Fe.shape[0], Fe.shape[1], Fe.shape[2], Ye.shape[-2]),dtype=np.float32)
        return Fy
    if Ye.ndim < 4: # ensure 4-d
        Ye = Ye.reshape((1,)*(4-Ye.ndim)+Ye.shape)    
    if dedup0 is not None: # remove duplicate copies output=0
        Ye = dedupY0(Ye)

    # inner-product score    
    Fy = np.einsum("mTEe,TEYe->mTEY", Fe, Ye, dtype=Fe.dtype)
    Fy = Fy.astype(Fe.dtype)

    # add correction for other measures
    if outputscore == 'sse':
        YR = convYR(Ye,R,offset) # (nM,nTrl,nSamp,nY,nFilt)
        # Apply the correction:
        #  SSE = (wX-Yr).^2
        #      = wX**2 - 2 wXYr + Yr**2
        #      = wX**2 = wX**2 - 2 Fe*Y + Yr**2
        #      = wX**2 - 2Fy + Yr**2
        #  take negative, so big (i.e. 0) is good, and drop constant over Y and divide by 2 ->
        #  -SSE = fY  = Fe*Y - .5* Yr**2
        Fy = Fy - np.sum(YR**2,-1) / 2

    elif outputscore == 'corr': # correlation based scoring...
        raise NotImplementedError()
    
    elif not outputscore == 'ip':
        raise NotImplementedError("output scoring with {} isn't supported".format(outputscore))
    
    return Fy

def dedupY0(Y, zerodup=True, yfeatdim=True):
    ''' remove outputs which are duplicates of the first (objID==0) output
    Inputs:
      Y=(tr,ep,Y,e)
      zerodup : bool
         if True, then if mi is the duplicate, then zero the  duplicate, i.e. of Y[...,mi,:]=0
         else, we zero out ojID==0, i.e. Y[...,0,:]=0
    Outputs:
      Y=(tr,ep,Y,e) version of Y with duplicates of 1st row of Y set to 0'''

    # record input shape so can get it back later
    Yshape = Y.shape
    #print("Y={}".format(Yshape))
    
    Y = np.copy(Y) # copy so can modify, w/o killing orginal
    # make the shape we want
    if not yfeatdim: # hack in feature dim
        Y = Y[..., np.newaxis]
    if Y.shape[-2] == 1:
        return Y.reshape(Yshape)
    if Y.ndim == 3: # add trial dim if not there
        Y = Y[np.newaxis, :, :, :]
        
    for ti in range(Y.shape[0]):
        # Note: horrible numpy hacks to make work & only for idx==0
        sim = np.sum(np.equal(Y[ti, :, 0:1, :], Y[ti, :, 1:, :]), axis=(0, 2))/(Y.shape[1]*Y.shape[3])
        #print("sim={}".format(sim))
        mi = np.argmax(sim)
        if sim[mi] > .95:
            #print("{}) dup {}={}".format(ti,0,mi+1))
            if zerodup: # zero out the duplicate of objId=0
                Y[ti, :, mi+1, :] = 0
            else: # zero out the objId==0 line
                Y[ti, :, 0, :] = 0
                
    # reshape back to input shape
    Y = np.reshape(Y, Yshape)
    return Y


def convWX(X,W):
    ''' apply spatial filter W  to X '''
    if W.ndim < 3:
        W=W.reshape((1,)*(3-W.ndim)+W.shape)
    if X.ndim < 3:
        X=X.reshape((1,)*(3-X.ndim)+X.shape)
    WX = np.einsum("TSd,mfd->mTSf",X,W)
    return WX #(nM,nTrl,nSamp,nfilt)

def convYR(Y,R,offset=0):
    ''' compute the convolution of Y with R '''
    if R is None:
        return Y
    if R.ndim < 4: # ensure 4-d
        R = R.reshape((1,)*(4-R.ndim)+R.shape) # (nM,nfilt,nE,tau)
    if Y.ndim < 4: # ensure 4-d
        Y = Y.reshape((1,)*(4-Y.ndim)+Y.shape) # (nTr,nSamp,nY,nE)
    #print("R={}".format(R.shape))
    #print("Y={}".format(Y.shape))
    
    # Compute the 'template' response
    Yt = window_axis(Y, winsz=R.shape[-1], axis=-3) # (nTr,nSamp,tau,nY,nE)
    #print("Yt={} (TStYe)".format(Yt.shape))
    #print("R={} (mfet)".format(R[...,::-1].shape))
    # TODO []: check treating filters correctly
    # TODO []: check if need to time-reverse IRF - A: YES!
    YtR = np.einsum("TStYe,mfet->mTSYf", Yt, R[...,::-1]) # (nM,nTr,nSamp,nY,nE)
    #print("YtR={} (mTSYf)".format(YtR.shape))
    # TODO []: correct edge effect correction, rather than zero-padding....
    # zero pad to keep the output size
    # TODO[]: pad with the *actual* values...
    tmp = np.zeros(YtR.shape[:-3]+(Y.shape[-3],)+YtR.shape[-2:])
    #print("tmp={}".format(tmp.shape))
    tmp[..., R.shape[-1]-1-offset:YtR.shape[-3]+R.shape[-1]-1-offset, :, :] = YtR
    YtR = tmp
    #print("YtR={}".format(YtR.shape))
    return YtR #(nM,nTrl,nSamp,nY,nfilt)

def convXYR(X,Y,W,R,offset):
    WX=convWX(X,W) # (nTr,nSamp,nfilt)
    YR=convYR(Y,R,offset) # (nTr,nSamp,nY,nfilt)
    # Sum out  filt dimesion
    WXYR = np.sum(WX[...,np.newaxis,:]*YR,-1) #(nTr,nSamp,nY)
    return WXYR,WX,YR

#@function
def testcases():
    #   Fe  = [nE x nEpoch x nTrl x nM ] similarity score for each event type for each stimulus
    #   Ye  = [nE x nY x nEpoch x nTrl ] Indicator for which events occured for which outputs
    nE=2
    nEpoch=100
    nTrl=30
    nY=20
    nM=20
    Fe=np.random.standard_normal((nM,nTrl,nEpoch,nE))
    Ye=np.random.standard_normal((nTrl,nEpoch,nY,nE))
    print("Fe={}".format(Fe.shape))
    print("Ye={}".format(Ye.shape))
    Fy=scoreOutput(Fe,Ye) # (nM,nTrl,nEp,nY)
    print("Fy={}".format(Fy.shape))
    import matplotlib.pyplot as plt
    sFy=np.cumsum(Fy,axis=-2)
    plt.clf();plt.plot(sFy[0,0,:,:]);plt.xlabel('epoch');plt.ylabel('output');plt.show()


    # more complex example with actual signal/noise
    from utils import testSignal
    from scoreOutput import scoreOutput, plot_outputscore, convWX, convYR, convXYR
    from scoreStimulus import scoreStimulus
    from decodingSupervised import decodingSupervised
    from normalizeOutputScores import normalizeOutputScores
    import numpy as np
    irf=(1,1,-1,-1,0,0,0,0,0,0)
    X,Y,st,W,R = testSignal(nTrl=1,nSamp=1000,d=1,nE=1,nY=10,isi=2,irf=irf,noise2signal=0)

    plot_outputscore(X[0,...],Y[0,:,0:3,:],W,R)
    
    # add a correlated output
    Y[:,:,1,:]=Y[:,:,0,:]*.5
    plot_outputscore(X[0,...],Y[0,:,0:3,:],W,R)
    

def datasettest():
    # N.B. imports in function to avoid import loop..
    from datasets import get_dataset
    from model_fitting import MultiCCA, BwdLinearRegression, FwdLinearRegression
    from analyse_datasets import debug_test_dataset
    from scoreOutput import plot_outputscore
    from decodingCurveSupervised import decodingCurveSupervised
    if True:
        tau_ms=300
        offset_ms=0
        rank=1
        evtlabs=None
        l,f,_=get_dataset('twofinger')
        oX,oY,coords=l(f[0])
        
    else:
        tau_ms=20
        offset_ms=0
        rank=8
        evtlabs=None
        l,f,_=get_dataset('mark_EMG')
        X,Y,coords=l(f[1],whiten=True,stopband=((0,10),(45,55),(200,-1)),filterbank=((10,20),(20,45),(55,95),(105,200)))
        oX=X.copy(); oY=Y.copy()

    X=oX.copy()
    Y=oY.copy()

    # test with reduced number  classes?
    nY=8
    Y=Y[:,:,:nY+1,:nY]

    plt.close('all')
    debug_test_dataset(X, Y, coords, tau_ms=tau_ms, offset_ms=offset_ms, evtlabs=evtlabs, rank=8, outputscore='ip', model='cca')
        
    fs = coords[1]['fs']
    tau = min(X.shape[-2],int(tau_ms*fs/1000))
    offset=int(offset_ms*fs/1000)
    cca = MultiCCA(tau=tau, offset=offset, rank=rank, evtlabs=evtlabs)
    cca.fit(X,Y)
    Fy = cca.predict(X, Y, dedup0=True)
    (_) = decodingCurveSupervised(Fy)
    W=cca.W_
    R=cca.R_
    b=cca.b_

    plot_outputscore(X[0,...],Y[0,...],W,R)

       
def plot_Fy(Fy,cumsum=True):
    import matplotlib.pyplot as plt
    import numpy as np
    '''plot the output score function'''
    if cumsum:
        Fy = np.cumsum(Fy.copy(),-2)
    plt.clf()
    nPlts=min(25,Fy.shape[0])
    if Fy.shape[0]/2 > nPlts:
        tis = np.linspace(0,Fy.shape[0]/2-1,nPlts,dtype=int)
    else:
        tis = np.arange(0,nPlts,dtype=int)
    ncols = int(np.ceil(np.sqrt(nPlts)))
    nrows = int(np.ceil(nPlts/ncols))
    #fig, plts = plt.subplots(nrows, ncols, sharex='all', sharey='all', squeeze=False)
    axploti = ncols*(nrows-1)
    ax = plt.subplot(nrows,ncols,axploti+1)
    for ci,ti in enumerate(tis):
        # make the axis
        if ci==axploti: # common axis plot
            pl = ax
        else: # normal plot
            pl = plt.subplot(nrows,ncols,ci+1,sharex=ax, sharey=ax) # share limits
            plt.tick_params(labelbottom=False,labelleft=False) # no labels        
        #pl = plts[ci//ncols, ci%ncols]
        pl.plot(Fy[ti,:,1:])
        pl.plot(Fy[ti,:,0:1],'k',linewidth=5)
        pl.set_title("{}".format(ti))
        pl.grid(True)
    pl.legend(range(Fy.shape[-1]-1))
    plt.suptitle('cumsum Fy')
    
def plot_Fycomparsion(Fy,Fys,ti=0):    
    import matplotlib.pyplot as plt
    import numpy as np
    from normalizeOutputScores import normalizeOutputScores
    from decodingSupervised import decodingSupervised
    plt.subplot(231); plt.cla(); plt.plot(Fy[ti,:,:],label='Fy');  plt.title('Inner-product'); 

    # Ptgt every sample, accum from trial start
    ssFy,varsFy,N,nEp,nY=normalizeOutputScores(Fy, minDecisLen=-1, filtLen=5)
    plt.subplot(232); plt.cla(); plt.plot(np.cumsum(Fy[ti,:,:],-2),label='sFy');plt.plot(varsFy[ti,1:],'k-',linewidth=5,label='scale'); plt.title('cumsum(Inner-product)');  plt.title('cumsum(Inner-product)');
    Yest,Perr,Ptgt,decismdl,decisEp=decodingSupervised(Fy, minDecisLen=-1, marginalizemodels=False, filtLen=5)
    plt.subplot(233); plt.cla(); plt.plot(Ptgt[ti,:,:],label='Ptgt');  plt.title('Ptgt');

    #X2 = np.sum(X**2,axis=-1,keepdims=True); Fys=Fys-X2/2     #  include norm of X to be sure.
    plt.subplot(234); plt.cla(); plt.plot(Fys[ti,:,:],label='sFy');  plt.title('-SSE'); 

    # Ptgt every sample, accum from trial start
    ssFy,varsFy,N,nEp,nY=normalizeOutputScores(Fys, minDecisLen=-1, filtLen=5)
    Yest,Perr,Ptgt,decismdl,decisEp=decodingSupervised(Fys, minDecisLen=-1, marginalizemodels=False, filtLen=5)
    plt.subplot(235); plt.cla(); plt.plot(np.cumsum(Fys[ti,:,:],-2),label='sFy'); plt.plot(varsFy[ti,1:],'k-',linewidth=5,label='scale'); plt.title('cumsum(-SSE)');
    plt.subplot(236); plt.cla(); plt.plot(Ptgt[ti,:,:],label='Ptgt');  plt.title('Ptgt(-SSE)'); 


def plot_outputscore(X,Y,W=None,R=None,offset=0):
    import matplotlib.pyplot as plt
    import numpy as np
    from model_fitting import MultiCCA
    from scoreStimulus import scoreStimulus
    from scoreOutput import scoreOutput
    from decodingCurveSupervised import decodingCurveSupervised
    ''' plot the factored model to visualize the fit and the scoring functions'''
    if W is None:
        cca = MultiCCA(tau=R,evtlabs=None)
        if X.ndim < 3: # BODGE:
            cca.fit(X[np.newaxis,...], Y[np.newaxis,...])
        else:
            cca.fit(X, Y)
        W=cca.W_
        R=cca.R_

        Fy=cca.predict(X,Y)
        (_) = decodingCurveSupervised(Fy)

        
    WXYR,WX,YR=convXYR(X,Y,W,R,offset)
    
    plt.clf();
    plt.subplot(511);plt.plot(np.squeeze(WX));plt.grid();plt.title("WX");
    plt.subplot(512);plt.plot(Y[...,0],label='Y');plt.plot(np.squeeze(R).T,'k',label='R',linewidth=5);plt.grid();plt.title("Y");
    plt.subplot(513);plt.plot(np.squeeze(YR[...,0]));plt.grid();plt.title("YR");
    
    plt.subplot(5,3,10);plt.plot(np.squeeze(WXYR));plt.grid();plt.title("WXYR")
    plt.subplot(5,3,11);plt.plot(np.squeeze(np.cumsum(WXYR,-2)));plt.grid();plt.title("cumsum(WXYR)")

    err = WX[...,np.newaxis,:]-YR
    sse = np.sum(err**2,-1)

    plt.subplot(5,3,12);plt.plot(np.squeeze(np.cumsum(-sse,-2)));plt.grid();plt.title("cumsum(sse)")

    cor = np.cumsum(WXYR,-2)/np.sqrt(np.cumsum(np.sum(YR**2,-1),-2))
    plt.subplot(5,3,12);plt.cla();plt.plot(np.squeeze(np.cumsum(-sse,-2)));plt.grid();plt.title("corr")

    Fe = scoreStimulus(X,W,R)
    Fy = scoreOutput(Fe,Y,R=R,outputscore='ip')
    Fys = scoreOutput(Fe,Y,R=R,outputscore='sse')

    plt.subplot(5,3,13);plt.plot(np.squeeze(Fy));plt.grid();plt.title("Fy")
    plt.subplot(5,3,14);plt.plot(np.squeeze(np.cumsum(Fy,-2)));plt.grid();plt.title("cumsum(Fy)")
    plt.subplot(5,3,15);plt.plot(np.squeeze(np.cumsum(2*Fys,-2)));plt.grid();plt.title("cumsum(Fy(sse))")
    
if __name__=="__main__":
    testcases()
